"""Module for Tsallis entropy estimator."""

from numpy import column_stack, issubdtype, integer

from ..base import EntropyEstimator
from ..utils.array import assure_2d_data
from ..utils.exponential_family import (
    calculate_common_entropy_components,
    exponential_family_iq,
    exponential_family_i1,
)
from ... import Config
from ...utils.types import LogBaseType


class TsallisEntropyEstimator(EntropyEstimator):
    r"""Estimator for the Tsallis entropy.

    Attributes
    ----------
    data : array-like
        The data used to estimate the entropy.
    k : int
        The number of nearest neighbors used in the estimation.
    q : float
        The Tsallis parameter, order or exponent.
        Sometimes denoted as :math:`q`, analogous to the Rényi parameter :math:`\alpha`.

    Raises
    ------
    ValueError
        If the Tsallis parameter is not a positive number.
    ValueError
        If the number of nearest neighbors is not a positive integer.

    Notes
    -----
    In the :math:`q \to 1` limit, the Jackson sum (q-additivity) reduces to
    ordinary summation,
    and the Tallis entropy reduces to Shannon Entropy.
    This class of entropy measure is in particularly useful in the study in connection
    with long-range correlated systems and with non-equilibrium phenomena.
    """

    def __init__(
        self,
        data,
        *,  # all following parameters are keyword-only
        k: int = 4,
        q: float | int = None,
        base: LogBaseType = Config.get("base"),
    ):
        r"""Initialize the TsallisMIEstimator.

        Parameters
        ----------
        k : int
            The number of nearest neighbors to consider.
        q : float | int
            The Tsallis parameter, order or exponent.
            Sometimes denoted as :math:`q`, analogous to the Rényi parameter :math:`\alpha`.
        """
        super().__init__(data, base)
        if not isinstance(q, (int, float)) or q <= 0:
            raise ValueError("The Tsallis parameter ``q`` must be a positive number.")
        if not issubdtype(type(k), integer) or k <= 0:
            raise ValueError(
                "The number of nearest neighbors must be a positive integer."
            )
        self.k = k
        self.q = q
        self.data = assure_2d_data(data)

    def _simple_entropy(self):
        """Calculate the entropy of the data.

        Returns
        -------
        float
            The Tsallis entropy.
        """
        V_m, rho_k, N, m = calculate_common_entropy_components(self.data, self.k)

        if self.q != 1:
            # Tsallis entropy for q != 1
            I_N_k_q = exponential_family_iq(self.k, self.q, V_m, rho_k, N, m)
            return (1 - I_N_k_q) / (self.q - 1)
        else:
            # Shannon entropy (limes for alpha = 1)
            return exponential_family_i1(self.k, V_m, rho_k, N, m, self._log_base)

    def _joint_entropy(self):
        """Calculate the joint Tsallis entropy of the data.

        This is done by joining the variables into one space
        and calculating the entropy.

        Returns
        -------
        float
            The joint Tsallis entropy.
        """
        self.data = column_stack(self.data)
        return self._simple_entropy()



"""Module for the Ordinal / Permutation transfer entropy estimator."""

from abc import ABC

from numpy import integer, issubdtype, ndarray

from ... import Config
from ...utils.config import logger
from ...utils.types import LogBaseType
from ..base import (
    ConditionalTransferEntropyEstimator,
    EffectiveValueMixin,
    PValueMixin,
    TransferEntropyEstimator,
)
from ..utils.discrete_transfer_entropy import combined_te_form
from ..utils.ordinal import symbolize_series
from ..utils.te_slicing import cte_observations, te_observations


class BaseOrdinalTEEstimator(ABC):
    r"""Base class for the Ordinal / Permutation transfer entropy.

    Attributes
    ----------
    source, dest : array-like
        The source (X) and dest (Y) data used to estimate the transfer entropy.
    cond : array-like, optional
        The conditional data used to estimate the conditional transfer entropy.
    embedding_dim : int
        The size of the permutation patterns.
    stable : bool, optional
        If True, when sorting the data, the order of equal elements is preserved.
        This can be useful for reproducibility and testing, but might be slower.
    prop_time : int, optional
        Number of positions to shift the data arrays relative to each other (multiple of
        ``step_size``).
        Delay/lag/shift between the variables, representing propagation time.
        Not compatible with the ``cond`` parameter / conditional TE.
        Alternatively called `offset`.
    step_size : int, optional
        Step size between elements for the state space reconstruction.
    src_hist_len, dest_hist_len : int, optional
        Number of past observations to consider for the source and destination data.
    cond_hist_len : int, optional
        Number of past observations to consider for the conditional data.
        Only used for conditional transfer entropy.

    Raises
    ------
    ValueError
        If the ``embedding_dim`` is negative or not an integer.
    ValueError
        If the ``embedding_dim`` is too large for the given data.
    ValueError
        If ``step_size``, ``prop_time``, and ``embedding_dim`` are such that
        the data is too small.
    TypeError
        If the data are not 1d array-like(s).

    Notes
    -----
    If ``embedding_dim`` is set to 1, the transfer entropy is always 0.
    """

    def __init__(
        self,
        source,
        dest,
        *,  # Enforce keyword-only arguments
        cond=None,
        embedding_dim: int,
        stable: bool = False,
        prop_time: int = 0,
        step_size: int = 1,
        src_hist_len: int = 1,
        dest_hist_len: int = 1,
        cond_hist_len: int = 1,
        offset: int = None,
        base: LogBaseType = Config.get("base"),
    ):
        """Initialize the BaseOrdinalTEEstimator.

        Parameters
        ----------
        source, dest : array-like
            The source (X) and destination (Y) data used to estimate the transfer entropy.
        cond : array-like, optional
            The conditional data used to estimate the conditional transfer entropy.
        embedding_dim : int
            The embedding dimension of the Ordinal entropy.
        stable : bool, optional
            If True, when sorting the data, the order of equal elements is preserved.
            This can be useful for reproducibility and testing, but might be slower.
        prop_time : int, optional
            Number of positions to shift the data arrays relative to each other (multiple of
            ``step_size``).
            Delay/lag/shift between the variables, representing propagation time.
            Assumed time taken by info to transfer from source to destination
            Not compatible with the ``cond`` parameter / conditional TE.
            Alternatively called `offset`.
        step_size : int, optional
            Step size between elements for the state space reconstruction.
        src_hist_len, dest_hist_len : int, optional
            Number of past observations to consider for the source and destination data.
        cond_hist_len : int, optional
            Number of past observations to consider for the conditional data.
            Only used for conditional transfer entropy.
        """
        self.source = None
        self.dest = None
        self.src_hist_len = None
        self.dest_hist_len = None
        self.step_size = None
        if cond is None:
            super().__init__(
                source,
                dest,
                prop_time=prop_time,
                src_hist_len=src_hist_len,
                dest_hist_len=dest_hist_len,
                step_size=step_size,
                offset=offset,
                base=base,
            )
        else:
            super().__init__(
                source,
                dest,
                cond=cond,
                step_size=step_size,
                src_hist_len=src_hist_len,
                dest_hist_len=dest_hist_len,
                cond_hist_len=cond_hist_len,
                prop_time=prop_time,
                offset=offset,
                base=base,
            )
            if self.cond.ndim > 1:
                raise TypeError(
                    "The conditional variable must be an 1d array, "
                    "so that ordinal patterns can be computed from it."
                )
        if not issubdtype(type(embedding_dim), integer) or embedding_dim < 0:
            raise ValueError("The embedding_dim must be a non-negative integer.")
        if embedding_dim == 1:
            logger.warning(
                "The Ordinal mutual information is always 0 for embedding_dim=1."
            )
        if any(var.ndim > 1 for var in (self.source, self.dest)):
            raise TypeError(
                "The data must be tuples of 1D arrays. "
                "Ordinal patterns can only be computed from 1D arrays."
            )
        if not issubdtype(type(step_size), integer) or step_size < 0:
            raise ValueError("The step_size must be a non-negative integer.")
        if len(self.source) < (embedding_dim - 1) * step_size + 1:
            raise ValueError(
                "The data is too small for the given step_size and embedding_dim."
            )
        self.embedding_dim = embedding_dim
        self.stable = stable


class OrdinalTEEstimator(
    BaseOrdinalTEEstimator, PValueMixin, EffectiveValueMixin, TransferEntropyEstimator
):
    r"""Estimator for the Ordinal / Permutation transfer entropy.

    Attributes
    ----------
    source, dest : array-like
        The source (X) and dest (Y) data used to estimate the transfer entropy.
    embedding_dim : int
        The size of the permutation patterns.
    prop_time : int, optional
        Number of positions to shift the data arrays relative to each other (multiple of
        ``step_size``).
        Delay/lag/shift between the variables, representing propagation time.
        Assumed time taken by info to transfer from source to destination.
        Alternatively called `offset`.
    step_size : int
        Step size between elements for the state space reconstruction.
    src_hist_len, dest_hist_len : int
        Number of past observations to consider for the source and destination data.

    Raises
    ------
    ValueError
        If the ``embedding_dim`` is negative or not an integer.
    ValueError
        If the ``embedding_dim`` is too large for the given data.
    ValueError
        If ``step_size``, ``prop_time``, and ``embedding_dim`` are such that
        the data is too small.

    Notes
    -----
    If ``embedding_dim`` is set to 1, the transfer entropy is always 0.
    """

    def _calculate(self) -> float:
        """Calculate the Ordinal / Permutation transfer entropy."""
        self.symbols = tuple(
            symbolize_series(
                var, self.embedding_dim, self.step_size, to_int=True, stable=self.stable
            )
            for var in (self.source, self.dest)
        )
        return combined_te_form(
            te_observations,
            *self.symbols,
            local=False,
            log_func=self._log_base,
            src_hist_len=self.src_hist_len,
            dest_hist_len=self.dest_hist_len,
            step_size=self.step_size,
            permute_src=self.permute_src,
            resample_src=self.resample_src,
        )

    def _extract_local_values(self) -> ndarray:
        """Separately calculate the local values.

        Returns
        -------
        ndarray[float]
            The calculated local values of cmi.
        """
        return combined_te_form(
            te_observations,
            *self.symbols,
            local=True,
            log_func=self._log_base,
            src_hist_len=self.src_hist_len,
            dest_hist_len=self.dest_hist_len,
            step_size=self.step_size,
            permute_src=self.permute_src,
            resample_src=self.resample_src,
        )


class OrdinalCTEEstimator(BaseOrdinalTEEstimator, ConditionalTransferEntropyEstimator):
    r"""Estimator for the Ordinal / Permutation conditional transfer entropy.

    Attributes
    ----------
    source, dest, cond : array-like
        The source (X), destination (Y), and conditional (Z) data used to estimate the
        conditional transfer entropy.
    embedding_dim : int
        The size of the permutation patterns.
    prop_time : int, optional
        Number of positions to shift the data arrays relative to each other (multiple of
        ``step_size``).
        Delay/lag/shift between the variables, representing propagation time.
        Assumed time taken by info to transfer from source to destination.
    step_size : int
        Step size between elements for the state space reconstruction.
    src_hist_len, dest_hist_len, cond_hist_len : int
        Number of past observations to consider for the source, destination, and conditional data.

    Raises
    ------
    ValueError
        If the ``embedding_dim`` is negative or not an integer.
    ValueError
        If the ``embedding_dim`` is too large for the given data.
    ValueError
        If ``step_size``, ``prop_time``, and ``embedding_dim`` are such that the
        data is too small.

    Notes
    -----
    If ``embedding_dim`` is set to 1, the transfer entropy is always 0.
    """

    def _calculate(self) -> float:
        """Calculate the Ordinal / Permutation conditional transfer entropy."""
        self.symbols = tuple(
            symbolize_series(
                var, self.embedding_dim, self.step_size, to_int=True, stable=self.stable
            )
            for var in (self.source, self.dest, self.cond)
        )
        return combined_te_form(
            cte_observations,
            *self.symbols,
            local=False,
            log_func=self._log_base,
            src_hist_len=self.src_hist_len,
            dest_hist_len=self.dest_hist_len,
            cond_hist_len=self.cond_hist_len,
            step_size=self.step_size,
        )

    def _extract_local_values(self) -> ndarray:
        """Separately calculate the local values.

        Returns
        -------
        ndarray[float]
            The calculated local values of cmi.
        """
        return combined_te_form(
            cte_observations,
            *self.symbols,
            local=True,
            log_func=self._log_base,
            src_hist_len=self.src_hist_len,
            dest_hist_len=self.dest_hist_len,
            cond_hist_len=self.cond_hist_len,
            step_size=self.step_size,
        )



"""Normalize a time series."""

from numpy import max as np_max
from numpy import min as np_min
from numpy import ndarray


def normalize_data_0_1(data: ndarray) -> ndarray:
    """
    Normalize the given data to scale each dimension to the range [0, 1].

    Parameters
    ----------
    data : 2D array-like
        List of data points where each row is a data point in d-dimensional space.

    Returns
    -------
    2D array
        Data scaled to the range [0, 1].

    Examples
    --------
    >>> data = np.array([[1, 2], [3, 4], [5, 6]])
    >>> normalize_data_0_1(data)
    array([[0. , 0. ],
           [0.5, 0.5],
           [1. , 1. ]])
    """
    min_vals = np_min(data, axis=0)
    max_vals = np_max(data, axis=0)
    return (data - min_vals) / (max_vals - min_vals)



"""Module for the kernel entropy estimator."""

from numpy import column_stack

from ... import Config
from ...utils.types import LogBaseType
from ..base import EntropyEstimator, WorkersMixin
from ..utils.array import assure_2d_data
from ..utils.kde import kde_probability_density_function


class KernelEntropyEstimator(WorkersMixin, EntropyEstimator):
    """Estimator for entropy (Shannon) using Kernel Density Estimation (KDE).

    Attributes
    ----------
    data : array-like
        The data used to estimate the entropy.
    bandwidth : float | int
        The bandwidth for the kernel.
    kernel : str
        Type of kernel to use, compatible with the KDE
        implementation :func:`kde_probability_density_function() <infomeasure.estimators.utils.kde.kde_probability_density_function>`.
    workers : int, optional
       Number of workers to use for parallel processing.
       Default is 1, meaning no parallel processing.
       If set to -1, all available CPU cores will be used.

    Notes
    -----
    A small ``bandwidth`` can lead to under-sampling,
    while a large ``bandwidth`` may over-smooth the data, obscuring details.
    """

    def __init__(
        self,
        data,
        *,  # all following parameters are keyword-only
        bandwidth: float | int,
        kernel: str,
        workers: int = 1,
        base: LogBaseType = Config.get("base"),
    ):
        """Initialize the KernelEntropyEstimator.

        Parameters
        ----------
        bandwidth : float | int
            The bandwidth for the kernel.
        kernel : str
            Type of kernel to use, compatible with the KDE
            implementation :func:`kde_probability_density_function() <infomeasure.estimators.utils.kde.kde_probability_density_function>`.
        workers : int, optional
           Number of workers to use for parallel processing.
           Default is 1, meaning no parallel processing.
           If set to -1, all available CPU cores will be used.
        """
        super().__init__(data, workers=workers, base=base)
        self.data = assure_2d_data(data)
        self.bandwidth = bandwidth
        self.kernel = kernel

    def _simple_entropy(self):
        """Calculate the entropy of the data.

        Returns
        -------
        array-like
            The local form of the entropy.
        """
        # Compute the KDE densities
        densities = kde_probability_density_function(
            self.data, self.bandwidth, kernel=self.kernel, workers=self.n_workers
        )

        # Compute the log of the densities
        return -self._log_base(densities)

    def _joint_entropy(self):
        """Calculate the joint entropy of the data.

        This is done by joining the variables into one space
        and calculating the entropy.

        Returns
        -------
        array-like
            The local form of the joint entropy.
        """
        self.data = column_stack(self.data)
        return self._simple_entropy()



"""Functions for interaction information, a multivariate generalization of
mutual information."""

from collections import Counter

from numpy import (
    clip,
    uint64,
    log,
    ndarray,
    ones,
    prod,
    ravel,
    unique,
    zeros,
)
from numpy import (
    sum as np_sum,
)
from scipy.sparse import find as sp_find
from scipy.stats.contingency import crosstab
from sparse import COO, asnumpy


def mutual_information_global(*data: tuple, log_func: callable = log) -> float:
    """Estimate the global mutual information between multiple random variables.

    Parameters
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the global mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
    log_func : callable, optional
        The logarithm function to use. Default is the natural logarithm.

    Returns
    -------
    float
        The global mutual information between the random variables.
    """
    if all(d.ndim == 1 for d in data):
        if len(data) == 2:
            return _mutual_information_global_2d_int(*data, log_func=log_func)
        else:
            return _mutual_information_global_nd_int(*data, log_func=log_func)
    else:
        return _mutual_information_global_nd_other(*data, log_func=log_func)


def _mutual_information_global_nd_int(*data: tuple, log_func: callable = log) -> float:
    """Estimate the global mutual information between an arbitrary number of
    random variables."""
    uniques, indices = zip(*[unique(var, return_inverse=True, axis=0) for var in data])
    contingency_coo = COO(
        coords=indices,
        data=ones(len(indices[0]), dtype=uint64),
        shape=tuple(len(uniq) for uniq in uniques),
        fill_value=0,
    )

    # Non-zero indices and values
    idxs = contingency_coo.nonzero()
    vals = contingency_coo.data

    # Marginal probabilities
    count_marginals = [
        asnumpy(
            contingency_coo.sum(  # all axes, except i
                axis=tuple(range(contingency_coo.ndim))[:i]
                + tuple(range(contingency_coo.ndim))[i + 1 :]
            )
        )
        for i in range(contingency_coo.ndim)
    ]

    # Early return if any of the marginal entropies is zero
    if any(count_m.size == 1 for count_m in count_marginals):
        return 0.0

    # Calculate the expected logarithm values for the outer product of marginal
    # probabilities, only for non-zero entries.
    outer = prod(
        [
            count.take(idx).astype(uint64, copy=False)
            for count, idx in zip(count_marginals, idxs)
        ],
        axis=0,
    )

    # Normalized contingency table (joint probability)
    contingency_sum = contingency_coo.sum()
    p_joint = vals / contingency_sum

    # Logarithm of the non-zero elements
    log_p_joint = log_func(vals)

    log_outer = -log_func(outer) + len(count_marginals) * log_func(contingency_sum)
    # Combine the terms to calculate the mutual information
    mi = p_joint * (log_p_joint - log_func(contingency_sum)) + p_joint * log_outer
    return mi.sum()  # interaction information can be negative, do not clip


def _mutual_information_global_2d_int(
    *data: tuple[ndarray[int]], log_func: callable = log
) -> float:
    """Estimate the global mutual information between two random variables.

    The approach relies on the contingency table of the two variables.
    Instead of calculating the full outer product, only the non-zero elements are
    considered.
    Code adapted from
    the :func:`mutual_info_score() <sklearn.metrics.mutual_info_score>` function in
    scikit-learn.

    Parameters
    ----------

    """
    # Contingency table - COOrdinate sparse matrix
    contingency_coo = crosstab(*data, sparse=True).count
    # Non-zero indices and values
    nzx, nzy, nzv = sp_find(contingency_coo)

    # Normalized contingency table (joint probability)
    contingency_sum = contingency_coo.sum()
    p_joint = nzv / contingency_sum
    # Marginal probabilities
    pi = ravel(contingency_coo.sum(axis=1))
    pj = ravel(contingency_coo.sum(axis=0))

    # Early return if any of the marginal entropies is zero
    if pi.size == 1 or pj.size == 1:
        return 0.0

    # Logarithm of the non-zero elements
    log_p_joint = log_func(nzv)

    # Calculate the expected logarithm values for the outer product of marginal
    # probabilities, only for non-zero entries.
    outer = pi.take(nzx).astype(uint64, copy=False) * pj.take(nzy).astype(
        uint64, copy=False
    )
    log_outer = -log_func(outer) + log_func(pi.sum()) + log_func(pj.sum())
    # Combine the terms to calculate the mutual information
    mi = p_joint * (log_p_joint - log_func(contingency_sum)) + p_joint * log_outer
    return clip(mi.sum(), 0.0, None)


def _mutual_information_global_nd_other(
    *data: tuple[ndarray], log_func: callable = log
) -> float:
    """Alternative method to estimate the global mutual information between an
    arbitrary number of random variables.

    Same as :func:`_mutual_information_global_nd_int`, but for non-integer data.
    """
    # data is a tuple of ndarrays, for joint data, concatenate these rowsj
    # joint_data = [tuple(row) for row in column_stack(data)]
    joint_data = [tuple(tuple(val) for val in row) for row in zip(*data)]

    # Count joint and marginal occurrences
    joint_counts = Counter(joint_data)
    joint_total = sum(joint_counts.values())
    marginal_counts = [Counter([tuple(val) for val in var]) for var in data]
    marginal_totals = [sum(counts.values()) for counts in marginal_counts]

    # Estimate probabilities
    joint_prob = {key: val / joint_total for key, val in joint_counts.items()}
    marginal_prob = [
        {key: val / total for key, val in counts.items()}
        for counts, total in zip(marginal_counts, marginal_totals)
    ]

    # Calculate the mutual information
    mi_sum = [
        joint_prob[key]
        * log_func(
            joint_prob[key]
            / prod([marginal_prob[i][key[i]] for i in range(len(data))], axis=0)
        )
        for key in joint_prob
    ]
    if len(mi_sum) == 0:
        return 0.0
    return np_sum(mi_sum)


def mutual_information_local(*data: tuple, log_func: callable = log) -> ndarray:
    """Estimate the local mutual information between multiple random variables.

    The mean of the local mutual information is the global mutual information.
    Only calculating the global value is more efficient,
    so evaluating the local mutual information should only be done
    when explicitly needed.

    Parameters
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the local mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
    log_func : callable, optional
        The logarithm function to use. Default is the natural logarithm.

    Returns
    -------
    ndarray
        The local mutual information between the random variables.
    """
    # Contingency table - COOrdinate sparse matrix
    uniques, indices = zip(*[unique(var, return_inverse=True, axis=0) for var in data])
    contingency_coo = COO(
        coords=indices,
        data=ones(len(indices[0]), dtype=uint64),
        shape=tuple(len(uniq) for uniq in uniques),
        fill_value=0,
    )

    # Normalized contingency table (joint probability)
    contingency_sum = contingency_coo.sum()
    # Marginal probabilities
    count_marginals = [
        asnumpy(
            contingency_coo.sum(  # all axes, except i
                axis=tuple(range(contingency_coo.ndim))[:i]
                + tuple(range(contingency_coo.ndim))[i + 1 :]
            )
        )
        for i in range(contingency_coo.ndim)
    ]

    # Early return if any of the marginal entropies is zero
    if any(count_m.size == 1 for count_m in count_marginals):
        return zeros(len(data[0]))

    # To get local values we iterate over *data
    # for each row in the input data: log( p(data) / p(data1) * p(data2) )
    p_joint = contingency_coo[indices].data / contingency_sum
    outer = prod(
        [count[indices[i]] for i, count in enumerate(count_marginals)]
        / contingency_sum,
        axis=0,
    )
    return -log_func(outer) + log_func(p_joint)


def conditional_mutual_information_global(
    *data: tuple, cond: ndarray, log_func: callable = log
) -> float:
    """Estimate the global conditional mutual information
    between multiple random variables and a conditioning variable.

    Parameters
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the global mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
    cond : array-like, shape (n_samples,)
        The conditioning variable.
    log_func : callable, optional
        The logarithm function to use. Default is the natural logarithm.

    Returns
    -------
    float
        The global conditional mutual information between the random variables.

    Raises
    ------
    ValueError
        If the conditioning variable is not one-dimensional.
    """
    if cond.ndim != 1:
        raise ValueError("The conditioning variable must be one-dimensional.")
    return _conditional_mutual_information_global_nd_int(
        *data, cond=cond, log_func=log_func
    )


def _conditional_mutual_information_global_nd_int(
    *data: tuple, cond: ndarray, log_func: callable = log
) -> float:
    """Estimate the global conditional mutual information between an arbitrary number of
    random variables and a conditioning variable."""
    uniques, indices = zip(
        *[unique(var, return_inverse=True, axis=0) for var in (data + (cond,))]
    )
    contingency_coo = COO(
        coords=indices,
        data=ones(len(indices[0]), dtype=uint64),
        shape=tuple(len(uniq) for uniq in uniques),
        fill_value=0,
    )

    # Non-zero indices and values
    idxs = contingency_coo.nonzero()
    vals = contingency_coo.data

    # Marginal-conditioned probabilities
    count_marginals_cond = [
        asnumpy(
            contingency_coo.sum(  # all axes, except i and cond
                axis=tuple(range(contingency_coo.ndim - 1))[:i]
                + tuple(range(contingency_coo.ndim - 1))[i + 1 :]
            )
        )
        for i in range(contingency_coo.ndim - 1)
    ]
    count_cond = asnumpy(
        contingency_coo.sum(axis=tuple(range(contingency_coo.ndim - 1)))
    )
    # all axes, except cond

    # Early return if any of the marginal entropies is zero
    if any(count_m.size == 1 for count_m in count_marginals_cond):
        return 0.0

    # Calculate the expected logarithm values for the outer product of marginal
    # probabilities, only for non-zero entries.
    outer = prod(
        [
            count[idx, idxs[-1]].astype(uint64, copy=False)
            for count, idx in zip(count_marginals_cond, idxs[:-1])
        ],
        axis=0,
    )

    # Normalized contingency table (joint probability)
    contingency_sum = contingency_coo.sum()
    p_joint = vals / contingency_sum

    # Logarithm of the non-zero elements
    p_cond = count_cond.take(idxs[-1]).astype(uint64, copy=False)
    log_p_joint = log_func(vals * p_cond) - log_func(contingency_sum)

    log_outer = -log_func(outer) + sum(
        log_func(count_m.sum()) for count_m in count_marginals_cond[:-1]
    )
    # Combine the terms to calculate the mutual information
    mi = p_joint * log_p_joint + p_joint * log_outer
    return mi.sum()  # interaction information can be negative, do not clip


def conditional_mutual_information_local(
    *data: tuple, cond: ndarray, log_func: callable = log
) -> ndarray:
    """Estimate the local conditional mutual information between multiple
    random variables and a conditioning variable.

    The mean of the local conditional mutual information is the
    global conditional mutual information.
    Only calculating the global value is more efficient,
    so evaluating the local conditional mutual information should only be done
    when explicitly needed.

    Parameters
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the local mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
    cond : array-like, shape (n_samples,)
        The conditioning variable.
    log_func : callable, optional
        The logarithm function to use. Default is the natural logarithm.

    Returns
    -------
    ndarray
        The local conditional mutual information between the random variables.
    """
    # Contingency table - COOrdinate sparse matrix
    uniques, indices = zip(
        *[unique(var, return_inverse=True, axis=0) for var in (data + (cond,))]
    )
    contingency_coo = COO(
        coords=indices,
        data=ones(len(indices[0]), dtype=uint64),
        shape=tuple(len(uniq) for uniq in uniques),
        fill_value=0,
    )

    # Normalized contingency table (joint probability)
    contingency_sum = contingency_coo.sum()
    # Marginal-conditioned probabilities
    count_marginals_cond = [
        asnumpy(
            contingency_coo.sum(  # all axes, except i and cond
                axis=tuple(range(contingency_coo.ndim - 1))[:i]
                + tuple(range(contingency_coo.ndim - 1))[i + 1 :]
            )
        )
        for i in range(contingency_coo.ndim - 1)
    ]
    count_cond = asnumpy(
        contingency_coo.sum(axis=tuple(range(contingency_coo.ndim - 1)))
    )

    # Early return if any of the marginal entropies is zero
    if any(count_m.size == 1 for count_m in count_marginals_cond):
        return zeros(len(data[0])).astype(float)

    # To get local values we iterate over *data
    # for each row in the input data: log( p(data) / p(data1) * p(data2) )
    p_joint = contingency_coo[indices].data / contingency_sum
    p_cond = count_cond[indices[-1]] / contingency_sum
    outer = prod(
        [count[indices[i], indices[-1]] for i, count in enumerate(count_marginals_cond)]
        / contingency_sum,
        axis=0,
    )
    return log_func(p_joint * p_cond) - log_func(outer)



"""Module for the discrete transfer entropy estimator."""

from abc import ABC

from numpy import ndarray

from ... import Config
from ...utils.config import logger
from ...utils.types import LogBaseType
from ..base import (
    ConditionalTransferEntropyEstimator,
    EffectiveValueMixin,
    PValueMixin,
    TransferEntropyEstimator,
)
from ..utils.discrete_transfer_entropy import combined_te_form
from ..utils.te_slicing import cte_observations, te_observations


class BaseDiscreteTEEstimator(ABC):
    """Base class for discrete transfer entropy estimators.

    Attributes
    ----------
    source, dest : array-like
        The source (X) and destination (Y) data used to estimate the transfer entropy.
    cond : array-like, optional
        The conditional data used to estimate the conditional transfer entropy.
    prop_time : int, optional
        Number of positions to shift the data arrays relative to each other (multiple of
        ``step_size``).
        Delay/lag/shift between the variables, representing propagation time.
        Assumed time taken by info to transfer from source to destination.
        Not compatible with the ``cond`` parameter / conditional TE.
        Alternatively called `offset`.
    step_size : int, optional
        Step size between elements for the state space reconstruction.
    src_hist_len, dest_hist_len : int, optional
        Number of past observations to consider for the source and destination data.
    cond_hist_len : int, optional
        Number of past observations to consider for the conditional data.
        Only used for conditional transfer entropy.
    """

    def __init__(
        self,
        source,
        dest,
        *,  # Enforce keyword-only arguments
        cond=None,
        prop_time: int = 0,
        step_size: int = 1,
        src_hist_len: int = 1,
        dest_hist_len: int = 1,
        cond_hist_len: int = 1,
        offset: int = None,
        base: LogBaseType = Config.get("base"),
    ):
        """Initialize the BaseDiscreteTEEstimator.

        Parameters
        ----------
        source, dest : array-like
            The source (X) and destination (Y) data used to estimate the transfer entropy.
        cond : array-like, optional
            The conditional data used to estimate the conditional transfer entropy.
        prop_time : int, optional
            Number of positions to shift the data arrays relative to each other (multiple of
            ``step_size``).
            Delay/lag/shift between the variables, representing propagation time.
            Assumed time taken by info to transfer from source to destination
            Not compatible with the ``cond`` parameter / conditional TE.
            Alternatively called `offset`.
        step_size : int, optional
            Step size between elements for the state space reconstruction.
        src_hist_len, dest_hist_len : int, optional
            Number of past observations to consider for the source and destination data.
        cond_hist_len : int, optional
            Number of past observations to consider for the conditional data.
            Only used for conditional transfer entropy.
        """
        self.source = source
        self.dest = dest
        if cond is None:
            super().__init__(
                source,
                dest,
                prop_time=prop_time,
                src_hist_len=src_hist_len,
                dest_hist_len=dest_hist_len,
                step_size=step_size,
                offset=offset,
                base=base,
            )
        else:
            super().__init__(
                source,
                dest,
                cond=cond,
                step_size=step_size,
                src_hist_len=src_hist_len,
                dest_hist_len=dest_hist_len,
                cond_hist_len=cond_hist_len,
                prop_time=prop_time,
                offset=offset,
                base=base,
            )
            if (
                self.source.dtype.kind == "f"
                or self.dest.dtype.kind == "f"
                or (self.cond is not None and self.cond.dtype.kind == "f")
            ):
                logger.warning(
                    "The data looks like a float array ("
                    f"source: {self.source.dtype}, dest: {self.dest.dtype}). "
                    "Make sure the data is properly symbolized or discretized "
                    "for the transfer entropy estimation."
                )
            if hasattr(self, "cond") and self.cond.dtype.kind == "f":
                logger.warning(
                    "The conditional data looks like a float array ("
                    f"{self.cond.dtype}). "
                    "Make sure the data is properly symbolized or discretized "
                    "for the conditional transfer entropy estimation."
                )


class DiscreteTEEstimator(
    BaseDiscreteTEEstimator, PValueMixin, EffectiveValueMixin, TransferEntropyEstimator
):
    """Estimator for discrete transfer entropy.

    Attributes
    ----------
    source, dest : array-like
        The source (X) and destination (Y) data used to estimate the transfer entropy.
    prop_time : int, optional
        Number of positions to shift the data arrays relative to each other (multiple of
        ``step_size``).
        Delay/lag/shift between the variables, representing propagation time.
        Assumed time taken by info to transfer from source to destination.
        Alternatively called `offset`.
    step_size : int
        Step size between elements for the state space reconstruction.
    src_hist_len, dest_hist_len : int
        Number of past observations to consider for the source and destination data.
    """

    def _calculate(self):
        """Estimate the Discrete Transfer Entropy."""
        return combined_te_form(
            te_observations,
            self.source,
            self.dest,
            local=False,
            log_func=self._log_base,
            src_hist_len=self.src_hist_len,
            dest_hist_len=self.dest_hist_len,
            step_size=self.step_size,
            permute_src=self.permute_src,
            resample_src=self.resample_src,
        )

    def _extract_local_values(self) -> ndarray:
        """Separately calculate the local values.

        Returns
        -------
        ndarray[float]
            The calculated local values of cmi.
        """
        return combined_te_form(
            te_observations,
            self.source,
            self.dest,
            local=True,
            log_func=self._log_base,
            src_hist_len=self.src_hist_len,
            dest_hist_len=self.dest_hist_len,
            step_size=self.step_size,
            permute_src=self.permute_src,
            resample_src=self.resample_src,
        )


class DiscreteCTEEstimator(
    BaseDiscreteTEEstimator, ConditionalTransferEntropyEstimator
):
    """Estimator for discrete conditional transfer entropy.

    Attributes
    ----------
    source, dest, cond : array-like
        The source (X), destination (Y), and conditional (Z) data used to estimate the
        conditional transfer entropy.
    step_size : int
        Step size between elements for the state space reconstruction.
    src_hist_len, dest_hist_len, cond_hist_len : int, optional
        Number of past observations to consider for the source, destination,
        and conditional data.
    prop_time : int, optional
        Not compatible with the ``cond`` parameter / conditional TE.
    """

    def _calculate(self):
        """Estimate the Discrete Transfer Entropy."""
        return combined_te_form(
            cte_observations,
            self.source,
            self.dest,
            self.cond,
            local=False,
            log_func=self._log_base,
            src_hist_len=self.src_hist_len,
            dest_hist_len=self.dest_hist_len,
            cond_hist_len=self.cond_hist_len,
            step_size=self.step_size,
        )

    def _extract_local_values(self) -> ndarray:
        """Separately calculate the local values.

        Returns
        -------
        ndarray[float]
            The calculated local values of cmi.
        """
        return combined_te_form(
            cte_observations,
            self.source,
            self.dest,
            self.cond,
            local=True,
            log_func=self._log_base,
            src_hist_len=self.src_hist_len,
            dest_hist_len=self.dest_hist_len,
            cond_hist_len=self.cond_hist_len,
            step_size=self.step_size,
        )



"""Jensen-Shannon Divergence (JSD)."""

from numpy import sum as np_sum, concatenate, ndarray

from ..estimators.base import DistributionMixin
from ..estimators.entropy import (
    RenyiEntropyEstimator,
    TsallisEntropyEstimator,
    KozachenkoLeonenkoEntropyEstimator,
    KernelEntropyEstimator,
)
from ..estimators.functional import get_estimator_class


def jensen_shannon_divergence(*data, approach: str | None = None, **kwargs):
    r"""Calculate the Jensen-Shannon Divergence between two or more distributions.

    The Jensen-Shannon Divergence is a symmetrized and smoothed version of the
    Kullback-Leibler Divergence. It is calculated as the average of the
    Kullback-Leibler Divergence between each distribution and the average
    distribution.

    .. math::

        JSD(P \| Q) = \frac{1}{2} KL(P \| M) + \frac{1}{2} KL(Q \| M)

    where :math:`M = \frac{1}{2} (P + Q)`.

    Parameters
    ----------
    p : array-like
        The first data.
    q : array-like
        The second data.
    ... : array-like
        Further data to compare.
    approach : str
        The name of the entropy estimator to use.
    **kwargs : dict
        Additional keyword arguments to pass to the entropy estimator.

    Returns
    -------
    float
        The Jensen-Shannon Divergence.

    Raises
    ------
    ValueError
        If the approach is not supported or the entropy estimator is not
        compatible with the Jensen-Shannon Divergence.
    ValueError
        If any of the given data is not an array-like object.
    """
    if approach is None:
        raise ValueError("The approach must be specified.")
    if not all(isinstance(var, (list, ndarray)) for var in data):
        raise ValueError("All data must be array-like objects.")
    estimator_class = get_estimator_class(measure="entropy", approach=approach)
    if issubclass(
        estimator_class,
        (
            RenyiEntropyEstimator,
            TsallisEntropyEstimator,
            KozachenkoLeonenkoEntropyEstimator,
        ),
    ):
        raise ValueError(
            "The Jensen-Shannon Divergence is not supported for the "
            f"{estimator_class.__name__} estimator."
        )
    # if estimator_class has mixin DistributionMixin
    # then we can use the distribution method
    if issubclass(estimator_class, DistributionMixin):
        estimators = tuple(estimator_class(var, **kwargs) for var in data)
        marginal = sum(estimator.global_val() for estimator in estimators) / len(data)
        # the distributions have some matching and some unique keys, create a new dict
        # with the sum of the values of union of keys
        dists = [estimator.distribution() for estimator in estimators]
        # dict(
        #   m_i: (p(x_i) + q(x_i) + ... + r(x_i)) / n
        # )
        dists = {
            key: sum(dist.get(key, 0) for dist in dists) / len(dists)
            for key in set().union(*dists)
        }
        mixture = list(dists.values())
        mixture = -np_sum(mixture * estimators[0]._log_base(mixture))
        return mixture - marginal
    if issubclass(estimator_class, KernelEntropyEstimator):
        # The mixture distribution is the union of the data, as the kernel density
        # estimation is applied afterward.
        mix_est = estimator_class(concatenate(data, axis=0), **kwargs)
        return mix_est.global_val() - sum(
            estimator_class(var, **kwargs).global_val() for var in data
        ) / len(data)
    else:
        raise ValueError(  # pragma: no cover
            f"The approach {approach} is not supported."
        )



"""Module for the discrete entropy estimator."""

from numpy import sum as np_sum, ndarray, unique

from ..base import EntropyEstimator, DistributionMixin
from ..utils.ordinal import reduce_joint_space
from ... import Config
from ...utils.config import logger
from ...utils.types import LogBaseType


class DiscreteEntropyEstimator(DistributionMixin, EntropyEstimator):
    """Estimator for discrete entropy (Shannon entropy).

    Attributes
    ----------
    data : array-like
        The data used to estimate the entropy.
    """

    def __init__(self, data, *, base: LogBaseType = Config.get("base")):
        """Initialize the DiscreteEntropyEstimator."""
        super().__init__(data, base=base)
        # warn if the data looks like a float array
        if isinstance(self.data, ndarray) and self.data.dtype.kind == "f":
            logger.warning(
                "The data looks like a float array ("
                f"{data.dtype}). "
                "Make sure it is properly symbolized or discretized "
                "for the entropy estimation."
            )
        elif isinstance(self.data, tuple) and any(
            isinstance(marginal, ndarray) and marginal.dtype.kind == "f"
            for marginal in self.data
        ):
            logger.warning(
                "Some of the data looks like a float array. "
                "Make sure it is properly symbolized or discretized "
                "for the entropy estimation."
            )
        if (isinstance(self.data, ndarray) and self.data.ndim > 1) or isinstance(
            self.data, tuple
        ):
            # As the discrete shannon entropy disregards the order of the data,
            # we can reduce the values to unique integers.
            # In case of having multiple random variables (tuple or list),
            # this enumerates the unique co-occurrences.
            self.data = reduce_joint_space(self.data)

    def _simple_entropy(self):
        """Calculate the entropy of the data.

        Returns
        -------
        float
            The calculated entropy.
        """
        uniq, counts = unique(self.data, return_counts=True)
        probabilities = counts / self.data.shape[0]  # normalize
        self.dist_dict = dict(
            zip(uniq, probabilities)
        )  # store the distribution for later
        # Calculate the entropy
        return -np_sum(probabilities * self._log_base(probabilities))

    def _joint_entropy(self):
        """Calculate the joint entropy of the data.

        Returns
        -------
        float
            The calculated joint entropy.
        """
        # The data has already been reduced to unique values of co-occurrences
        return self._simple_entropy()

    def _extract_local_values(self):
        """Separately calculate the local values.

        Returns
        -------
        ndarray[float]
            The calculated local values of entropy.
        """
        p_local = [self.dist_dict[val] for val in self.data]
        return -self._log_base(p_local)



"""Ordinal / Permutation utility functions."""

from math import factorial
from typing import Generator

from numpy import (
    ndarray,
    argsort,
    apply_along_axis,
    zeros,
    sum as np_sum,
    array,
    uint64,
    iinfo,
    unique,
    column_stack,
)
from numpy.lib.stride_tricks import as_strided


def permutation_to_integer(perm: ndarray, dtype: type = uint64) -> int:
    """
    Convert a permutation pattern to a unique integer.
    The Lehmer code is used to convert the permutation to an integer.

    Parameters
    ----------
    perm : ndarray
        A permutation pattern.
    embedding_dim : int, optional
        The size of the permutation pattern. Default is
        None, which uses the length of the permutation.
        Using this, the maximal number will be set and the smnalles possible
        dtype will be used.

    Returns
    -------
    int : int, uint8, uint16, uint32, uint64
        A unique integer representing the permutation pattern.

    Examples
    --------
    >>> permutation_to_integer(array([0, 1]))
    0
    >>> permutation_to_integer(array([1, 0]))
    1
    >>> permutation_to_integer(array([0, 1, 2]))
    0
    >>> permutation_to_integer(array([2, 1, 0]))
    5

    Notes
    -----
    This approach has at least been known since
    1888 :cite:p:`laisantNumerationFactorielleApplication1888`.
    It is named after Derrick Henry Lehmer :cite:p:`Lehmer1960TeachingCT`.

    Raises
    ------
    ValueError
        If the embedding_dim is too large to convert to an uint64 (maximal 20).
    """
    n = len(perm)
    if n > 20:
        raise ValueError(
            "For embedding dimensions larger than 20, "
            "the integer will be too large for uint64."
        )
    factoradic = zeros(n, dtype=dtype)
    for i in range(n):
        factoradic[i] = np_sum(perm[i] > perm[i + 1 :], dtype=dtype)
    integer = np_sum(
        factoradic * array([factorial(n - 1 - i) for i in range(n)]),
        dtype=dtype,
    )
    return integer


def symbolize_series(
    series: ndarray, embedding_dim: int, step_size: int = 1, to_int=False, stable=False
) -> ndarray:
    """
    Convert a time series into a sequence of symbols (permutation patterns).

    Parameters
    ----------
    series : ndarray, shape (n,)
        A numpy array of data points.
    embedding_dim : int
        The size of the permutation patterns.
    step_size : int
        The step size for the sliding windows. Takes every `step_size`-th element.
    to_int : bool, optional
        Whether to convert the permutation patterns to integers. Default is False.
        This
    stable : bool, optional
        If True, when sorting the data, the embedding_dim of equal elements is preserved.
        This can be useful for reproducibility and testing, but might be slower.

    Returns
    -------
    patterns : ndarray, shape (n - (embedding_dim - 1) * step_size, embedding_dim)
        A list of tuples representing the symbolized series.

    Examples
    --------
    >>> series = np.array([1, 2, 3, 2, 1])
    >>> symbolize_series(series, 2, 1)
    array([[0, 1],
           [0, 1],
           [1, 0],
           [1, 0]])

    Raises
    ------
    ValueError
        If the embedding_dim is less than 1.
    ValueError
        If the step_size is less than 1.
    """
    if embedding_dim < 1:
        raise ValueError("The embedding_dim must be a positive integer.")
    if step_size < 1:
        raise ValueError("The step_size must be a positive integer.")
    # Create a view of the series with the given embedding_dim and step size
    shape = (series.size - (embedding_dim - 1) * step_size, embedding_dim)
    strides = (series.strides[0], series.strides[0] * step_size)
    # Extract subsequences
    subsequences = as_strided(series, shape=shape, strides=strides)
    # Get the permutation patterns
    patterns = apply_along_axis(argsort, 1, subsequences, stable=stable)

    # If Lehmer code is requested, convert the permutation to an integer
    if to_int:
        # Determine necessary dtype for the maximal number
        dtypes = ["uint8", "uint16", "uint32"]
        dtype = uint64
        for d in dtypes:  # try small to large
            if factorial(embedding_dim) < iinfo(d).max:
                dtype = d
                break
        # Convert the permutation patterns to integers
        patterns = apply_along_axis(
            lambda x: permutation_to_integer(x, dtype=dtype), 1, patterns
        )

    return patterns


def reduce_joint_space(data: ndarray | tuple[ndarray] | Generator) -> ndarray:
    """Reduce the data to the joint space.

    Reduce features: Assigns each unique feature vector to a unique integer.
    This is equivalent to the unique indices of the unique rows.
    If `data` is 1D, returns the data as is.

    Parameters
    ----------
    data : ndarray, shape (n_samples,) or (n_samples, n_features), or tuple of arrays
        The data to reduce.

    Returns
    -------
    ndarray, shape (n_samples,)
        The data in the joint space.

    Examples
    --------
    >>> from numpy import array
    >>> data = array([[1, 2], [2, 3], [1, 2], [2, 3], [3, 4]])
    >>> data.shape
    (5, 2)
    >>> reduce_joint_space(data)
    array([0, 1, 0, 1, 2])
    >>> reduce_joint_space(array([4, 5, 4, 5, 6]))
    array([4, 5, 4, 5, 6])
    >>> reduce_joint_space(array([[True, True, False], [False, True, False]]))
    array([0, 1])
    >>> reduce_joint_space(array([[3, 3], [2, 2], [1, 1]]))
    array([2, 1, 0])
    >>> reduce_joint_space((array([3, 2, 1]), array([3, 2, 1])))
    array([2, 1, 0])

    Raises
    ------
    ValueError
        If the data array is not 1D or 2D.
    ValueError
        If the data is a list.

    Notes
    -----
    The order of the unique values is not guaranteed, only the unique indices.
    This is because numpy sorts the values internally.
    """
    if isinstance(data, list):
        raise ValueError("Data array must be a numpy array or tuple of arrays.")
    if isinstance(data, Generator):
        data = tuple(data)
    if isinstance(data, tuple):
        data = column_stack(data)
    if data.ndim == 2:
        # Reduce the data to the joint space using the numpy unique function
        return unique(data, axis=0, return_inverse=True)[1]  # use inverse indices
    elif data.ndim == 1:
        return data
    raise ValueError("Data array must be 1D or 2D.")



"""Entropy estimators."""

from .discrete import DiscreteEntropyEstimator
from .kernel import KernelEntropyEstimator
from .kozachenko_leonenko import KozachenkoLeonenkoEntropyEstimator
from .renyi import RenyiEntropyEstimator
from .ordinal import OrdinalEntropyEstimator
from .tsallis import TsallisEntropyEstimator

__all__ = [
    "DiscreteEntropyEstimator",
    "KernelEntropyEstimator",
    "KozachenkoLeonenkoEntropyEstimator",
    "OrdinalEntropyEstimator",
    "RenyiEntropyEstimator",
    "TsallisEntropyEstimator",
]



"""Kernel Density Estimation (KDE) utilities."""

from multiprocessing import Pool, cpu_count

from numpy import (
    argsort,
    array_split,
    concatenate,
    cov,
    dot,
    inf,
    issubdtype,
    number,
)
from numpy import sum as np_sum
from numpy.linalg import eig
from scipy.spatial import KDTree
from scipy.stats import gaussian_kde

from ...utils.config import logger


def kde_probability_density_function(data, bandwidth, kernel="box", workers=-1):
    """
    Estimate the probability density function for a given data set using
    Kernel Density Estimation (KDE).

    Parameters
    ----------
    data : array
        A numpy array of data points, where each column represents a dimension.
    bandwidth : float
        The bandwidth for the kernel.
    kernel : str
        Type of kernel to use ('gaussian' or 'box').
    workers : int
        Number of parallel processes to use.
        -1: Use all available CPU cores.
        Default is 1.

    Returns
    -------
    ndarray[float]
        KDE at the given point(s).

    Raises
    ------
    ValueError
        If the kernel type is not supported
    ValueError
        If the bandwidth is not a positive number.
    """
    logger.debug(
        f"Called kde_probability_density_function with "
        f"kernel: {kernel}, workers: {workers}"
    )
    if not issubdtype(type(bandwidth), number) or bandwidth <= 0:
        raise ValueError("The bandwidth must be a positive number.")

    if kernel == "gaussian":
        if workers == -1:
            workers = cpu_count()
        return gaussian_kernel_densities(data.T, bandwidth, workers=workers)
    elif kernel == "box":
        # Get the number of data points (N) and the number of dimensions (d)
        N, d = data.shape

        # Calculate the volume of the box kernel
        volume = bandwidth**d

        logger.debug("Creating KDTree from data points for box KDE... ")
        tree = KDTree(data)
        logger.debug(
            f"KDTree created with {N} data points and {d} dimensions. "
            f"Querying KDTree for ball points with {workers} workers."
        )
        counts = tree.query_ball_point(
            data,
            bandwidth / 2,
            p=inf,
            return_length=True,
            workers=workers,
        )
        densities = counts / (N * volume)

        # Squeeze the densities array to remove any single-dimensional entries
        return densities.squeeze()
    # Another approach to box kernel density estimation
    else:  # TODO: Add more kernel types as needed
        raise ValueError(f"Unsupported kernel type: {kernel}. Use 'gaussian' or 'box'.")


def gaussian_kernel_densities(
    data, bandwidth, workers=1, eigen_threshold: float = 1e-10
):
    """Calculate kde for gaussian kernel.

    In case of multivariate data, checks rank of data and reduces dimensions
    if eigenvalues are below threshold.
    If already full rank, does no reprojection.

    Parameters
    ----------
    data : ndarray, shape (d, N)
        Data points to estimate density for.
    bandwidth : float
        Bandwidth parameter for kernel density estimation.
    workers : int, optional
        Number of workers to use for parallel processing. Default is 1.
    eigen_threshold : float, optional
        Threshold for eigenvalues to determine rank of data. Default is 1e-10.

    Returns
    -------
    densities : ndarray, shape (n,)
        Estimated density values at data points.
    """
    logger.debug(f"Calculating Gaussian KDE with bandwidth {bandwidth}...")
    if data.shape[0] > 1:  # Multivariate case
        # Calculate covariance matrix
        covariance_matrix = cov(data)
        # Get eigenvalues and eigenvectors
        values, vectors = eig(covariance_matrix)
        sorted_indices = argsort(values)[::-1]
        values_sorted = values[sorted_indices]
        vectors_sorted = vectors[:, sorted_indices]
        # Get the number of eigenvalues greater than the threshold
        num_non_zero_eigenvalues = np_sum(values_sorted > eigen_threshold)
        # Check projection necessary
        if num_non_zero_eigenvalues < data.shape[0]:
            logger.debug(
                f"Reducing dimensionality from {data.shape[1]} to "
                f"{num_non_zero_eigenvalues} dimensions."
            )
            # Project the data onto the reduced space
            pca_components = vectors_sorted[:, :num_non_zero_eigenvalues]
            data_projected = dot(data.T, pca_components).T
            logger.debug("Reprojected data, make kde...")
            # each worker get a chunk of data and evaluate KDE on it
            return parallel_kde_evaluate(data_projected, bandwidth, workers)

    return parallel_kde_evaluate(data, bandwidth, workers)


def query_chunk(params):
    """Evaluate KDE on a chunk of data."""
    full_data, query_data, bandwidth = params
    kde = gaussian_kde(full_data, bw_method=bandwidth)
    return kde.evaluate(query_data).squeeze()


def parallel_kde_evaluate(data, bandwidth, workers):
    """Evaluate KDE on a set of data in parallel.

    Parameters
    ----------
    data : array-like
        The data to evaluate the KDE on.
    bandwidth : float or str
       The bandwidth to use for the KDE.
    workers : int
        The number of worker processes to use for evaluation.

    Notes
    -----
    If the data is < 100000 samples or the number of workers is 1,
    evaluate the KDE on a single worker.
    """
    # parallelization just gets really effective, if chunk size is not too small
    # if data is not too large, use less workers than possible
    if workers == 1 or data.shape[1] < 20000:
        logger.debug(f"Evaluating kde on a single worker with data size {data.shape}.")
        kde = gaussian_kde(data, bw_method=bandwidth)
        return kde.evaluate(data).squeeze()

    workers = min(workers, max(1, data.shape[1] // 8000))
    query_chunks = array_split(data, workers, axis=1)
    logger.debug(
        f"Evaluating kde on {workers} workers with data size {data.shape} and "
        f"query chunks shape: {query_chunks[0].shape}."
    )
    pool = Pool(processes=workers)
    results = pool.map(
        query_chunk,
        zip([data] * len(query_chunks), query_chunks, [bandwidth] * len(query_chunks)),
    )
    pool.close()
    pool.join()
    return concatenate(results, axis=0)



"""infomeasure package."""

# Expose most common functions
from ._version import __version__
from .utils import Config
from .estimators.functional import (
    entropy,
    mutual_information,
    conditional_mutual_information,
    transfer_entropy,
    conditional_transfer_entropy,
    estimator,
    get_estimator_class,
)
from .composite_measures import jensen_shannon_divergence, kullback_leiber_divergence

h, mi, te = entropy, mutual_information, transfer_entropy
cmi, cte = conditional_mutual_information, conditional_transfer_entropy
jsd, kld = jensen_shannon_divergence, kullback_leiber_divergence

# Set package attributes
__author__ = "Carlson Büth"

__all__ = [
    "__version__",
    "__author__",
    "Config",
    "entropy",
    "mutual_information",
    "conditional_mutual_information",
    "transfer_entropy",
    "conditional_transfer_entropy",
    "estimator",
    "get_estimator_class",
    "jensen_shannon_divergence",
    "kullback_leiber_divergence",
    # aliases
    "h",
    "mi",
    "cmi",
    "te",
    "cte",
    "jsd",
    "kld",
]



"""Module for the Renyi mutual information estimator."""

from abc import ABC

from numpy import issubdtype, integer

from ..base import (
    PValueMixin,
    MutualInformationEstimator,
    ConditionalMutualInformationEstimator,
)
from ..entropy.renyi import RenyiEntropyEstimator
from ... import Config
from ...utils.types import LogBaseType


class BaseRenyiMIEstimator(ABC):
    r"""Base class for Renyi mutual information estimators.

    Attributes
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the (conditional) mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
        For conditional mutual information, only two data arrays are allowed.
    cond : array-like, optional
        The conditional data used to estimate the conditional mutual information.
    k : int
        The number of nearest neighbors used in the estimation.
    alpha : float | int
        The Rényi parameter, order or exponent.
        Sometimes denoted as :math:`\alpha` or :math:`q`.
    noise_level : float
        The standard deviation of the Gaussian noise to add to the data to avoid
        issues with zero distances.
    offset : int, optional
        Number of positions to shift the data arrays relative to each other.
        Delay/lag/shift between the variables. Default is no shift.
        Not compatible with the ``cond`` parameter / conditional MI.
    normalize : bool, optional
        If True, normalize the data before analysis.
    """

    def __init__(
        self,
        *data,
        cond=None,
        k: int = 4,
        alpha: float | int = None,
        noise_level=1e-8,
        offset: int = 0,
        normalize: bool = False,
        base: LogBaseType = Config.get("base"),
    ):
        r"""Initialize the estimator with specific parameters.

        Parameters
        ----------
        *data : array-like, shape (n_samples,)
            The data used to estimate the (conditional) mutual information.
            You can pass an arbitrary number of data arrays as positional arguments.
            For conditional mutual information, only two data arrays are allowed.
        cond : array-like, optional
            The conditional data used to estimate the conditional mutual information.
        k : int
            The number of nearest neighbors to consider.
        alpha : float | int
            The Renyi parameter, order or exponent.
            Sometimes denoted as :math:`\alpha` or :math:`q`.
        noise_level : float
            The standard deviation of the Gaussian noise to add to the data to avoid
            issues with zero distances.
        normalize
            If True, normalize the data before analysis.
        offset : int, optional
            Number of positions to shift the X and Y data arrays relative to each other.
            Delay/lag/shift between the variables. Default is no shift.
            Not compatible with the ``cond`` parameter / conditional MI.

        Raises
        ------
        ValueError
            If the Renyi parameter is not a positive number.
        ValueError
            If the number of nearest neighbors is not a positive integer.
        """
        if cond is None:
            super().__init__(*data, offset=offset, normalize=normalize, base=base)
        else:
            super().__init__(
                *data, cond=cond, offset=offset, normalize=normalize, base=base
            )
        if not isinstance(alpha, (int, float)) or alpha <= 0:
            raise ValueError("The Renyi parameter must be a positive number.")
        if not issubdtype(type(k), integer) or k <= 0:
            raise ValueError(
                "The number of nearest neighbors must be a positive integer."
            )
        self.k = k
        self.alpha = alpha
        self.noise_level = noise_level


class RenyiMIEstimator(BaseRenyiMIEstimator, PValueMixin, MutualInformationEstimator):
    r"""Estimator for the Renyi mutual information.

    Attributes
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
    k : int
        The number of nearest neighbors used in the estimation.
    alpha : float | int
        The Rényi parameter, order or exponent.
        Sometimes denoted as :math:`\alpha` or :math:`q`.
    noise_level : float
        The standard deviation of the Gaussian noise to add to the data to avoid
        issues with zero distances.
    offset : int, optional
        Number of positions to shift the data arrays relative to each other.
        Delay/lag/shift between the variables. Default is no shift.
    normalize : bool, optional
        If True, normalize the data before analysis.

    Notes
    -----
    The Rényi entropy is a generalization of Shannon entropy,
    where the small values of probabilities are emphasized for :math:`\alpha < 1`,
    and higher probabilities are emphasized for :math:`\alpha > 1`.
    For :math:`\alpha = 1`, it reduces to Shannon entropy.
    The Rényi-Entropy class can be particularly interesting for systems where additivity
    (in Shannon sense) is not always preserved, especially in nonlinear complex systems,
    such as when dealing with long-range forces.
    """

    def _calculate(self):
        """Calculate the mutual information of the data.

        Returns
        -------
        float
            Renyi mutual information of the data.
        """

        return self._generic_mi_from_entropy(
            estimator=RenyiEntropyEstimator,
            noise_level=self.noise_level,
            kwargs={"alpha": self.alpha, "k": self.k, "base": self.base},
        )


class RenyiCMIEstimator(BaseRenyiMIEstimator, ConditionalMutualInformationEstimator):
    r"""Estimator for the conditional Renyi mutual information.

    Attributes
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the conditional mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
    cond : array-like
        The conditional data used to estimate the conditional mutual information.
    k : int
        The number of nearest neighbors used in the estimation.
    alpha : float | int
        The Rényi parameter, order or exponent.
        Sometimes denoted as :math:`\alpha` or :math:`q`.
    noise_level : float
        The standard deviation of the Gaussian noise to add to the data to avoid
        issues with zero distances.
    offset : int, optional
        Number of positions to shift the data arrays relative to each other.
        Delay/lag/shift between the variables. Default is no shift.
    normalize : bool, optional
        If True, normalize the data before analysis.

    Notes
    -----
    The Rényi entropy is a generalization of Shannon entropy,
    where the small values of probabilities are emphasized for :math:`\alpha < 1`,
    and higher probabilities are emphasized for :math:`\alpha > 1`.
    For :math:`\alpha = 1`, it reduces to Shannon entropy.
    The Rényi-Entropy class can be particularly interesting for systems where additivity
    (in Shannon sense) is not always preserved, especially in nonlinear complex systems,
    such as when dealing with long-range forces.
    """

    def _calculate(self):
        """Calculate the conditional mutual information of the data.

        Returns
        -------
        float
            Conditional Renyi mutual information of the data.
        """
        return self._generic_cmi_from_entropy(
            estimator=RenyiEntropyEstimator,
            noise_level=self.noise_level,
            kwargs=dict(k=self.k, alpha=self.alpha, base=self.base),
        )



r"""Generalized data slicing method for transfer entropy estimators.

This module provides a method to slice the data arrays to prepare for transfer
entropy (TE) calculation.
The TE measures the information flow from a source variable (X) to
a target/destination variable (Y).
In this context, the future state is always associated with
the target/destination variable.

Conventions:

- ``X``: Source variable
- ``Y``: Destination/target variable
- ``dest_future``: Future state of the destination variable (Y)
- ``dest_history``: Past states of the destination variable (Y)
- ``src_history``: Past states of the source variable (X)

The TE is calculated as:

.. math::

    \hat{T}(Y_{t+1}|Y^{(k)}, X^{(l)}) = \frac{1}{N} \sum_{i=1}^{N} \log \frac{g(\hat{y}_{i+1}, y_i^{(k)}, x_i^{(l)}) g(\hat y_i^{(k)})}{g(y_i^{(k)}, x_i^{(l)}) g(\hat{y}_{i+1}, y_i^{(k)})}

"""

from typing import Iterable

from numpy import arange, ndarray, concatenate, expand_dims, issubdtype, integer
from numpy.random import Generator, default_rng

from ...utils.config import logger


def te_observations(
    source,
    destination,
    src_hist_len=1,
    dest_hist_len=1,
    step_size=1,
    permute_src=False,
    resample_src=False,
    construct_joint_spaces: bool = True,
) -> tuple[ndarray, ndarray, ndarray, ndarray] | Iterable | tuple:
    r"""
    Slice the data arrays to prepare for TE calculation.

    For TE there are four observations that are required to calculate the
    transfer entropy.

    .. math::

            \hat{T}(Y_{t+1}|Y^{(k)}, X^{(l)}) = \frac{1}{N} \sum_{i=1}^{N} \log \frac{g(\hat{y}_{i+1}, y_i^{(k)}, x_i^{(l)}) g(\hat y_i^{(k)})}{g(y_i^{(k)}, x_i^{(l)}) g(\hat{y}_{i+1}, y_i^{(k)})}

    Parameters
    ----------
    source : array, shape (n,)
        A numpy array of data points for the source variable (X).
    destination : array, shape (n,)
        A numpy array of data points for the destination variable (Y).
    src_hist_len : int, optional
        Number of past observations (l) to consider for the source data (X).
        Default is 1, only one current observation, no further history.
        One future observation is always considered for the source data.
    dest_hist_len : int, optional
        Number of past observations (k) to consider for the destination data (Y).
        Default is 1, only one current observation, no further history.
    step_size : int, optional
        Step size for the time delay in the embedding.
        Default is None, which equals to 1, every observation is considered.
        If step_size is greater than 1, the history is subsampled.
        This applies to both the source and destination data.
    permute_src : bool | Generator, optional
        Whether to shuffle the sliced source history data. Default is False.
        This is used for the permutation TE. Rows are permuted, keeping the
        history intact.
        If a random number generator is provided, it will be used for shuffling.
        If True, a new random number generator will be created.
    resample_src : bool | Generator, optional
        Whether to resample the sliced source history data. Default is False.
        This is used for the permutation TE using bootstrapping.
        Rows are resampled with replacement, keeping the history intact.
        If a random number generator is provided, it will be used for resampling.
        If True, a new random number generator will be created.
    construct_joint_spaces : bool, optional
        Whether to construct the joint spaces. Default is True.
        If False, the sliced source and destination data are returned instead.

    Returns
    -------
    joint_space_data : array, shape (max_len, src_hist_len + dest_hist_len + 1)
        :math:`g(x_i^{(l)}, y_i^{(k)}, \hat{y}_{i+1})`: Joint space data.
    dest_past_embedded : array, shape (max_len, dest_hist_len)
        :math:`g(\hat y_i^{(k)})` : Embedded past destination data.
    marginal_1_space_data : array, shape (max_len, dest_hist_len + src_hist_len)
        :math:`g(x_i^{(l)}, y_i^{(k)})` : Marginal space data for destination and
        source.
    marginal_2_space_data : array, shape (max_len, dest_hist_len + 1)
        :math:`g(y_i^{(k)}, \hat{y}_{i+1})` : Marginal space data for destination.
    sliced data : tuple of arrays
        If ``construct_joint_spaces`` is False, the sliced source and destination
        data are returned instead. Namely, the tuple contains:

        - ``src_history`` : array, shape (max_len, src_hist_len)
          :math:`x_i^{(l)}` : Source history.
        - ``dest_history`` : array, shape (max_len, dest_hist_len)
          :math:`y_i^{(k)}` : Destination history.
        - ``dest_future`` : array, shape (max_len,)
          :math:`\hat{y}_{i+1}` : Destination future.

    Notes
    -----
    - For permutation TE, ``permute_src`` xor ``resample_src`` can be used.
    - With ``max_len = data_len - (max(src_hist_len, dest_hist_len) - 1) * step_size``.

    Raises
    ------
    ValueError
        If the history (``src_hist_len`` or ``dest_hist_len`` times ``step_size``) is
        greater than the length of the data.
    ValueError
        If ``src_hist_len``, ``dest_hist_len``, or ``step_size`` are
        not positive integers.
    ValueError
        If both ``permute_src`` and ``resample_src`` are provided.
    """
    # log warning if step_size is >1 while src_hist_len or dest_hist_len are both 1
    if step_size > 1 and src_hist_len == 1 and dest_hist_len == 1:
        logger.warning(
            "If both ``src_hist_len`` and ``dest_hist_len`` are 1, "
            "having ``step_size`` > 1 does not impact the TE calculation."
        )
    # error if vars are not positive integers
    if not all(
        issubdtype(type(var), integer) and var > 0
        for var in (src_hist_len, dest_hist_len, step_size)
    ):
        raise ValueError(
            "src_hist_len, dest_hist_len, and step_size must be positive integers."
        )

    max_delay = max(dest_hist_len, src_hist_len) * step_size
    # max delay must be less than the length of the data, otherwise raise an error
    if max_delay >= len(source):
        raise ValueError(
            "The history demanded by the source and destination data "
            "is greater than the length of the data and results in empty arrays."
        )

    if permute_src and resample_src:
        raise ValueError("Only one of permute_src or resample_src can be provided.")

    base_indices = arange(max_delay, len(destination), step_size)

    # Construct src_history
    offset_indices = arange(step_size, (src_hist_len + 1) * step_size, step_size)
    src_history_indices = base_indices[:, None] - offset_indices[::-1]
    if isinstance(permute_src, Generator):
        permute_src.shuffle(src_history_indices, axis=0)  # in-place
    elif permute_src:
        rng = default_rng()
        rng.shuffle(src_history_indices, axis=0)  # in-place
    if isinstance(resample_src, Generator):
        src_history_indices = resample_src.choice(  # re-assign
            src_history_indices, size=src_history_indices.shape[0], axis=0, replace=True
        )
    elif resample_src:
        rng = default_rng()
        src_history_indices = rng.choice(  # re-assign
            src_history_indices, size=src_history_indices.shape[0], axis=0, replace=True
        )
    # get the data using the indices
    src_history = source[src_history_indices]

    # Construct dest_history
    offset_indices = arange(step_size, (dest_hist_len + 1) * step_size, step_size)
    dest_history_indices = base_indices[:, None] - offset_indices[::-1]
    dest_history = destination[dest_history_indices]

    # Construct dest_future
    dest_future = destination[base_indices]
    # src_future: (data_len,) -> (data_len, 1); or (data_len, m) -> (data_len, 1, m)
    dest_future = expand_dims(dest_future, axis=1)

    # flatten into two dimensions if the array has more than two dimensions
    src_history = (
        src_history
        if src_history.ndim < 3
        else src_history.reshape(src_history.shape[0], -1)
    )
    dest_history = (
        dest_history
        if dest_history.ndim < 3
        else dest_history.reshape(dest_history.shape[0], -1)
    )
    dest_future = (
        dest_future
        if dest_future.ndim < 3
        else dest_future.reshape(dest_future.shape[0], -1)
    )
    if construct_joint_spaces:
        return _construct_joint_space_data(
            src_history,  # x_i^{(l)}
            dest_history,  # y_i^{(k)}
            dest_future,  # \hat{y}_{i+1}
        )
    else:
        return src_history, dest_history, dest_future


def cte_observations(
    source,
    destination,
    condition,
    src_hist_len=1,
    dest_hist_len=1,
    cond_hist_len=1,
    step_size=1,
    construct_joint_spaces: bool = True,
) -> tuple[ndarray, ndarray, ndarray, ndarray] | Iterable | tuple:
    r"""
    Slice the data arrays to prepare for CTE calculation.

    For CTE there are four observations that are required to calculate the
    conditional transfer entropy.

    .. math::

            \hat{T}(Y_{t+1}|Y^{(k)}, X^{(l)}) = \frac{1}{N} \sum_{i=1}^{N} \log \frac{g(\hat{y}_{i+1}, y_i^{(k)}, z_i^{(m)}, x_i^{(l)}) g(\hat y_i^{(k)}, z_i^{(m)})}{g(y_i^{(k)}, z_i^{(m)}, x_i^{(l)}) g(\hat{y}_{i+1}, y_i^{(k)}, z_i^{(m)})}

    Parameters
    ----------
    source : array, shape (n,)
        A numpy array of data points for the source variable (X).
    destination : array, shape (n,)
        A numpy array of data points for the destination variable (Y).
    condition : array, shape (n,)
        A numpy array of data points for the conditioning variable (Z).
    src_hist_len : int, optional
        Number of past observations (l) to consider for the source data (X).
        Default is 1, only one current observation, no further history.
        One future observation is always considered for the source data.
    dest_hist_len : int, optional
        Number of past observations (k) to consider for the destination data (Y).
        Default is 1, only one current observation, no further history.
    cond_hist_len : int, optional
        Number of past observations (m) to consider for the conditioning data (Z).
        Default is 1, only one current observation, no further history.
    step_size : int, optional
        Step size for the time delay in the embedding.
        Default is None, which equals to 1, every observation is considered.
        If step_size is greater than 1, the history is subsampled.
        This applies to both the source and destination data.
    construct_joint_spaces : bool, optional
        Whether to construct the joint spaces. Default is True.
        If False, the sliced source and destination data are returned instead.

    Returns
    -------
    joint_space_data : array, shape (max_len, src_hist_len + dest_hist_len + cond_hist_len + 1)
        :math:`g(x_i^{(l)}, z_i^{(m)}, y_i^{(k)}, \hat{y}_{i+1})`:
        Conditional joint space data.
    dest_past_embedded : array, shape (max_len, dest_hist_len)
        :math:`g(\hat y_i^{(k)}, z_i^{(m)})` :
        Conditional embedded past destination data.
    marginal_1_space_data : array, shape (max_len, dest_hist_len + src_hist_len)
        :math:`g(x_i^{(l)}, z_i^{(m)}, y_i^{(k)})` :
        Conditional marginal space data for destination and source.
    marginal_2_space_data : array, shape (max_len, dest_hist_len + 1)
        :math:`g(z_i^{(m)}, y_i^{(k)}, \hat{y}_{i+1})` :
        Conditional marginal space data for destination.
    sliced data : tuple of arrays
        If ``construct_joint_spaces`` is False, the sliced source, destination and
        conditional data are returned instead.
        Namely, the tuple contains:

        - ``src_history`` : array, shape (max_len, src_hist_len)
          :math:`x_i^{(l)}` : Source history.
        - ``dest_history`` : array, shape (max_len, dest_hist_len)
          :math:`y_i^{(k)}` : Destination history.
        - ``dest_future`` : array, shape (max_len,)
          :math:`\hat{y}_{i+1}` : Destination future.
        - ``cond_history`` : array, shape (max_len, cond_hist_len)
          :math:`z_i^{(m)}` : Condition history.

    Notes
    -----
    With ``max_len = data_len - (max(src_hist_len, dest_hist_len, cond_hist_len) - 1)
    * step_size``.

    Raises
    ------
    TypeError
        If the arguments are wrong types.
    ValueError
        If the history (``src_hist_len`` or ``dest_hist_len`` or ``cond_hist_len`` times
        ``step_size``) is greater than the length of the data.
    ValueError
        If ``src_hist_len``, ``dest_hist_len``, ``cond_hist_len``, or ``step_size`` are
        not positive integers.
    """
    if not all(
        issubdtype(type(var), integer) and var > 0
        for var in (src_hist_len, dest_hist_len, cond_hist_len, step_size)
    ):
        raise ValueError(
            "src_hist_len, dest_hist_len, cond_hist_len, and step_size "
            "must be positive integers."
        )
    if not all(isinstance(var, ndarray) for var in (source, destination, condition)):
        raise TypeError("source, destination, and condition must be numpy arrays.")
    # log warning if step_size is >1 while src_hist_len or dest_hist_len are both 1
    if (
        step_size > 1
        and src_hist_len == 1
        and dest_hist_len == 1
        and cond_hist_len == 1
    ):
        logger.warning(
            "If all ``src_hist_len``, ``dest_hist_len``, and ``cond_hist_len`` are 1, "
            "having ``step_size`` > 1 does not impact the TE calculation."
        )
    # error if vars are not positive integers
    if not all(
        issubdtype(type(var), integer) and var > 0
        for var in (src_hist_len, dest_hist_len, cond_hist_len, step_size)
    ):
        raise ValueError(
            "src_hist_len, dest_hist_len, cond_hist_len, and step_size "
            "must be positive integers."
        )

    max_delay = max(dest_hist_len, src_hist_len, cond_hist_len) * step_size
    # max delay must be less than the length of the data, otherwise raise an error
    if max_delay >= len(source):
        raise ValueError(
            "The history demanded by the source, destination, and condition data "
            "is greater than the length of the data and results in empty arrays."
        )

    base_indices = arange(max_delay, len(destination), step_size)

    # Construct src_history
    offset_indices = arange(step_size, (src_hist_len + 1) * step_size, step_size)
    src_history_indices = base_indices[:, None] - offset_indices[::-1]
    src_history = source[src_history_indices]

    # Construct cond_history
    offset_indices = arange(step_size, (cond_hist_len + 1) * step_size, step_size)
    cond_history_indices = base_indices[:, None] - offset_indices[::-1]
    cond_history = condition[cond_history_indices]

    # Construct dest_history
    offset_indices = arange(step_size, (dest_hist_len + 1) * step_size, step_size)
    dest_history_indices = base_indices[:, None] - offset_indices[::-1]
    dest_history = destination[dest_history_indices]

    # Construct dest_future
    dest_future = destination[base_indices]
    # src_future: (data_len,) -> (data_len, 1); or (data_len, m) -> (data_len, 1, m)
    dest_future = expand_dims(dest_future, axis=1)

    # flatten into two dimensions if the array has more than two dimensions
    src_history = (
        src_history
        if src_history.ndim < 3
        else src_history.reshape(src_history.shape[0], -1)
    )
    cond_history = (
        cond_history
        if cond_history.ndim < 3
        else cond_history.reshape(cond_history.shape[0], -1)
    )
    dest_history = (
        dest_history
        if dest_history.ndim < 3
        else dest_history.reshape(dest_history.shape[0], -1)
    )
    dest_future = (
        dest_future
        if dest_future.ndim < 3
        else dest_future.reshape(dest_future.shape[0], -1)
    )

    if construct_joint_spaces:
        return _construct_joint_space_data(
            src_history,  # x_i^{(l)}
            dest_history,  # y_i^{(k)}
            dest_future,  # \hat{y}_{i+1}
            cond_history,  # z_i^{(m)}
        )
    else:
        return src_history, dest_history, dest_future, cond_history


def _construct_joint_space_data(
    *sliced_data,
) -> tuple[ndarray, ndarray, ndarray, ndarray]:
    """Construct joint space data.

    Parameters
    ----------
    src_history : np.ndarray
        Source history.
    dest_history : np.ndarray
        Destination history.
    dest_future : np.ndarray
        Destination future.
    cond_history : np.ndarray, optional
        Conditioning history.
    Returns
    -------
    tuple of np.ndarray
        Joint space data, as returned by :func:`te_observations` and
        :func:`cte_observations`, if conditioning history is provided.
    """

    # g(x_i^{(l)}, z_i^{(m)}, y_i^{(k)}, \hat{y}_{i+1})
    joint_space_data = concatenate(
        sliced_data,  # x_i^{(l)}, y_i^{(k)}, \hat{y}_{i+1}, z_i^{(m)}
        axis=1,  # or x_i^{(l)}, y_i^{(k)}, \hat{y}_{i+1}
    )

    if len(sliced_data) == 3:  # g(\hat y_i^{(k)})
        dest_past_embedded = sliced_data[1]
    else:  # g(\hat y_i^{(k)}, z_i^{(m)})
        dest_past_embedded = concatenate((sliced_data[1], *sliced_data[3:]), axis=1)

    # g(x_i^{(l)}, y_i^{(k)}, z_i^{(m)})
    marginal_1_space_data = concatenate(
        (*sliced_data[:2], *sliced_data[3:]), axis=1
    )  # x_i^{(l)}, y_i^{(k)}, z_i^{(m)} or x_i^{(l)}, y_i^{(k)}

    # g(y_i^{(k)}, \hat{y}_{i+1}, z_i^{(m)})
    marginal_2_space_data = concatenate(
        sliced_data[1:],  # y_i^{(k)}, \hat{y}_{i+1}, z_i^{(m)}
        axis=1,  # or y_i^{(k)}, \hat{y}_{i+1}
    )

    return (
        joint_space_data,
        dest_past_embedded,
        marginal_1_space_data,
        marginal_2_space_data,
    )



"""Module for the Kozacenko-Leonenko entropy estimator."""

from numpy import column_stack
from numpy import inf, log, issubdtype, integer
from scipy.spatial import KDTree
from scipy.special import digamma

from ..base import EntropyEstimator, RandomGeneratorMixin
from ..utils.array import assure_2d_data
from ..utils.unit_ball_volume import unit_ball_volume
from ... import Config
from ...utils.types import LogBaseType


class KozachenkoLeonenkoEntropyEstimator(RandomGeneratorMixin, EntropyEstimator):
    r"""Kozachenko-Leonenko estimator for Shannon entropies.

    Attributes
    ----------
    data : array-like
        The data used to estimate the entropy.
    k : int
        The number of nearest neighbors to consider.
    noise_level : float
        The standard deviation of the Gaussian noise to add to the data to avoid
        issues with zero distances.
    minkowski_p : float, :math:`1 \leq p \leq \infty`
        The power parameter for the Minkowski metric.
        Default is np.inf for maximum norm. Use 2 for Euclidean distance.

    Raises
    ------
    ValueError
        If the number of nearest neighbors is not a positive integer
    ValueError
        If the noise level is negative
    ValueError
        If the Minkowski power parameter is invalid

    Notes
    -----
    Changing the number of nearest neighbors ``k`` can change the outcome,
    but the default value of :math:`k=4` is recommended by :cite:p:`miKSG2004`.
    """

    def __init__(
        self,
        data,
        *,  # all following parameters are keyword-only
        k: int = 4,
        noise_level=1e-10,
        minkowski_p=inf,
        base: LogBaseType = Config.get("base"),
    ):
        r"""Initialize the Kozachenko-Leonenko estimator.

        Parameters
        ----------
        k : int
            The number of nearest neighbors to consider.
        noise_level : float
            The standard deviation of the Gaussian noise to add to the data to avoid
            issues with zero distances.
        minkowski_p : float, :math:`1 \leq p \leq \infty`
            The power parameter for the Minkowski metric.
            Default is np.inf for maximum norm. Use 2 for Euclidean distance.
        """
        if not issubdtype(type(k), integer) or k <= 0:
            raise ValueError(
                "The number of nearest neighbors (k) must be a positive "
                f"integer, but got {k}."
            )
        if noise_level < 0:
            raise ValueError(
                f"The noise level must be non-negative, but got {noise_level}."
            )
        if not (1 <= minkowski_p <= inf):
            raise ValueError(
                "The Minkowski power parameter must be positive, "
                f"but got {minkowski_p}."
            )
        super().__init__(data, base=base)
        self.data = assure_2d_data(data)
        self.k = k
        self.noise_level = noise_level
        self.minkowski_p = minkowski_p

    def _simple_entropy(self):
        """Calculate the entropy of the data.

        Returns
        -------
        float
            The calculated entropy.
        """
        # Copy the data to avoid modifying the original
        data_noisy = self.data.astype(float).copy()
        # Add small Gaussian noise to data to avoid issues with zero distances
        if self.noise_level and self.noise_level != 0:
            data_noisy += self.rng.normal(0, self.noise_level, self.data.shape)

        # Build a KDTree for efficient nearest neighbor search with maximum norm
        tree = KDTree(data_noisy)

        # Find the k-th nearest neighbors for each point
        distances, _ = tree.query(data_noisy, self.k + 1, p=self.minkowski_p)
        # Only keep the k-th nearest neighbor distance
        distances = distances[:, -1]

        # Constants for the entropy formula
        N = self.data.shape[0]
        d = self.data.shape[1]
        # Volume of the d-dimensional unit ball for maximum norm
        c_d = unit_ball_volume(d, r=1 / 2, p=self.minkowski_p)

        # Compute the local entropies
        local_h = -digamma(self.k) + digamma(N) + log(c_d) + d * log(2 * distances)
        # return in desired base
        return local_h / log(self.base) if self.base != "e" else local_h

    def _joint_entropy(self):
        """Calculate the joint entropy of the data.

        This is done by joining the variables into one space
        and calculating the entropy.

        Returns
        -------
        float
            The calculated joint entropy.
        """
        self.data = column_stack(self.data)
        return self._simple_entropy()



"""Mutual information estimators."""

from .discrete import DiscreteMIEstimator, DiscreteCMIEstimator
from .kernel import KernelMIEstimator, KernelCMIEstimator
from .kraskov_stoegbauer_grassberger import KSGMIEstimator, KSGCMIEstimator
from .renyi import RenyiMIEstimator, RenyiCMIEstimator
from .ordinal import OrdinalMIEstimator, OrdinalCMIEstimator
from .tsallis import TsallisMIEstimator, TsallisCMIEstimator

__all__ = [
    "DiscreteMIEstimator",
    "DiscreteCMIEstimator",
    "KernelMIEstimator",
    "KernelCMIEstimator",
    "KSGMIEstimator",
    "KSGCMIEstimator",
    "RenyiMIEstimator",
    "RenyiCMIEstimator",
    "OrdinalMIEstimator",
    "OrdinalCMIEstimator",
    "TsallisMIEstimator",
    "TsallisCMIEstimator",
]



"""Transfer entropy estimators."""

from .discrete import DiscreteTEEstimator, DiscreteCTEEstimator
from .kernel import KernelTEEstimator, KernelCTEEstimator
from .kraskov_stoegbauer_grassberger import KSGTEEstimator, KSGCTEEstimator
from .renyi import RenyiTEEstimator, RenyiCTEEstimator
from .ordinal import OrdinalTEEstimator, OrdinalCTEEstimator
from .tsallis import TsallisTEEstimator, TsallisCTEEstimator

__all__ = [
    "DiscreteTEEstimator",
    "DiscreteCTEEstimator",
    "KernelTEEstimator",
    "KernelCTEEstimator",
    "KSGTEEstimator",
    "KSGCTEEstimator",
    "RenyiTEEstimator",
    "RenyiCTEEstimator",
    "OrdinalTEEstimator",
    "OrdinalCTEEstimator",
    "TsallisTEEstimator",
    "TsallisCTEEstimator",
]



"""Module for the Kraskov-Stoegbauer-Grassberger (KSG) mutual information estimator."""

from abc import ABC

from numpy import column_stack, inf, array, ndarray, log
from scipy.spatial import KDTree
from scipy.special import digamma

from ..base import (
    MutualInformationEstimator,
    ConditionalMutualInformationEstimator,
    PValueMixin,
)
from ..utils.array import assure_2d_data
from ... import Config
from ...utils.types import LogBaseType


class BaseKSGMIEstimator(ABC):
    r"""Base class for mutual information using the Kraskov-Stoegbauer-Grassberger (KSG)
    method.

    Attributes
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the (conditional) mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
        For conditional mutual information, only two data arrays are allowed.
    cond : array-like, optional
        The conditional data used to estimate the conditional mutual information.
    k : int
        The number of nearest neighbors to consider.
    noise_level : float
        The standard deviation of the Gaussian noise to add to the data to avoid
        issues with zero distances.
    minkowski_p : float, :math:`1 \leq p \leq \infty`
        The power parameter for the Minkowski metric.
        Default is np.inf for maximum norm. Use 2 for Euclidean distance.
    offset : int, optional
        Number of positions to shift the data arrays relative to each other.
        Delay/lag/shift between the variables. Default is no shift.
        Not compatible with the ``cond`` parameter / conditional MI.
    normalize : bool, optional
        If True, normalize the data before analysis.
    """

    def __init__(
        self,
        *data,
        cond=None,
        k: int = 4,
        noise_level=1e-10,
        minkowski_p=inf,
        offset: int = 0,
        normalize: bool = False,
        base: LogBaseType = Config.get("base"),
    ):
        r"""Initialize the estimator with specific parameters.

        Parameters
        ----------
        *data : array-like, shape (n_samples,)
            The data used to estimate the (conditional) mutual information.
            You can pass an arbitrary number of data arrays as positional arguments.
            For conditional mutual information, only two data arrays are allowed.
        cond : array-like, optional
            The conditional data used to estimate the conditional mutual information.
        k : int
            The number of nearest neighbors to consider.
        noise_level : float
            The standard deviation of the Gaussian noise to add to the data to avoid
            issues with zero distances.
        minkowski_p : float, :math:`1 \leq p \leq \infty`
            The power parameter for the Minkowski metric.
            Default is np.inf for maximum norm. Use 2 for Euclidean distance.
        normalize
            If True, normalize the data before analysis.
        offset : int, optional
            Number of positions to shift the data arrays relative to each other.
            Delay/lag/shift between the variables. Default is no shift.
            Not compatible with the ``cond`` parameter / conditional MI.
        """
        self.data: tuple[ndarray] = None
        self.cond = None
        if cond is None:
            super().__init__(*data, offset=offset, normalize=normalize, base=base)
        else:
            super().__init__(
                *data, cond=cond, offset=offset, normalize=normalize, base=base
            )
            # Ensure self.cond is a 2D array
            self.cond = assure_2d_data(self.cond)
        self.k = k
        self.noise_level = noise_level
        self.minkowski_p = minkowski_p
        # Ensure self.data_x and self.data_y are 2D arrays
        self.data = tuple(assure_2d_data(var) for var in self.data)


class KSGMIEstimator(BaseKSGMIEstimator, PValueMixin, MutualInformationEstimator):
    r"""Estimator for mutual information using the Kraskov-Stoegbauer-Grassberger (KSG)
    method.

    Attributes
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
    k : int
        The number of nearest neighbors to consider.
    noise_level : float
        The standard deviation of the Gaussian noise to add to the data to avoid
        issues with zero distances.
    minkowski_p : float, :math:`1 \leq p \leq \infty`
        The power parameter for the Minkowski metric.
        Default is np.inf for maximum norm. Use 2 for Euclidean distance.
    offset : int, optional
        Number of positions to shift the data arrays relative to each other.
        Delay/lag/shift between the variables. Default is no shift.
    normalize : bool, optional
        If True, normalize the data before analysis.

    Notes
    -----
    Changing the number of nearest neighbors ``k`` can change the outcome,
    but the default value of :math:`k=4` is recommended by :cite:p:`miKSG2004`.
    """

    def _calculate(self) -> ndarray:
        """Calculate the mutual information of the data.

        Returns
        -------
        local_mi : array
            Local mutual information for each point.
        """
        # Copy the data to avoid modifying the original
        data = [var.astype(float).copy() for var in self.data]

        # Add Gaussian noise to the data if the flag is set
        if self.noise_level and self.noise_level != 0:
            for i in range(len(data)):
                for j in range(len(data[i])):
                    data[i][j] += self.rng.normal(0, self.noise_level, data[i][j].shape)

        # Stack the X and Y data to form joint observations
        data_joint = column_stack(data)

        # Create a KDTree for joint data to find nearest neighbors using the maximum
        # norm
        tree_joint = KDTree(data_joint)  # default leafsize is 10

        # Find the k-th nearest neighbor distance for each point in joint space using
        # the maximum norm
        distances, _ = tree_joint.query(data_joint, k=self.k + 1, p=self.minkowski_p)
        kth_distances = distances[:, -1]

        # Create KDTree objects for X and Y to count neighbors in marginal spaces using
        # the maximum norm
        trees_marginal = [KDTree(var) for var in data]

        # Count neighbors within k-th nearest neighbor distance in X and Y spaces using
        # the maximum norm
        counts_marginal = [
            [
                tree.query_ball_point(p, r=d, p=self.minkowski_p, return_length=True)
                - 1
                for p, d in zip(var, kth_distances)
            ]
            for tree, var in zip(trees_marginal, data)
        ]

        # Compute mutual information using the KSG estimator formula
        N = len(data[0])
        m = len(data)  # number of variables
        # Compute local mutual information for each point
        local_mi = array(
            [
                digamma(self.k)
                - sum(digamma(n + 1) for n in counts)
                + (m - 1) * digamma(N)
                for counts in zip(*counts_marginal)
            ]
        )

        return local_mi / log(self.base) if self.base != "e" else local_mi


class KSGCMIEstimator(BaseKSGMIEstimator, ConditionalMutualInformationEstimator):
    r"""Estimator for conditional mutual information using
    the Kraskov-Stoegbauer-Grassberger (KSG) method.

    Attributes
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the conditional mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
    cond : array-like
        The conditional data used to estimate the conditional mutual information.
    k : int
        The number of nearest neighbors to consider.
    noise_level : float
        The standard deviation of the Gaussian noise to add to the data to avoid
        issues with zero distances.
    minkowski_p : float, :math:`1 \leq p \leq \infty`
        The power parameter for the Minkowski metric.
        Default is np.inf for maximum norm. Use 2 for Euclidean distance.
    normalize : bool, optional
        If True, normalize the data before analysis.

    Notes
    -----
    Changing the number of nearest neighbors ``k`` can change the outcome,
    but the default value of :math:`k=4` is recommended by :cite:p:`miKSG2004`.
    """

    def _calculate(self) -> ndarray:
        """Calculate the conditional mutual information of the data.

        Returns
        -------
        local_cmi : array
            Local conditional mutual information for each point.
        """
        # Copy the data to avoid modifying the original
        data = [var.astype(float).copy() for var in self.data]
        cond = self.cond.astype(float).copy()

        # Add Gaussian noise to the data if the flag is set
        if self.noise_level and self.noise_level != 0:
            for j in range(len(data[0])):
                for i in range(len(data)):
                    data[i][j] += self.rng.normal(0, self.noise_level, data[i][j].shape)
                cond += self.rng.normal(0, self.noise_level, cond.shape)

        # Stack the X, Y, and Z data to form joint observations
        data_joint = column_stack((*data, cond))

        # Create KDTree for efficient nearest neighbor search in joint space
        tree_joint = KDTree(data_joint)

        # Find k-th nearest neighbor distances in joint space
        distances, _ = tree_joint.query(data_joint, k=self.k + 1, p=self.minkowski_p)
        kth_distances = distances[:, -1]

        # Count points within k-th nearest neighbor distance in marginal spaces
        trees_marginal_cond = [KDTree(column_stack((var, cond))) for var in data]
        tree_cond = KDTree(cond)

        # Count neighbors within k-th nearest neighbor distance in X and Y spaces using
        # the maximum norm
        counts_marginal_cond = [
            [
                tree.query_ball_point(p, r=d, p=self.minkowski_p, return_length=True)
                - 1
                for p, d in zip(column_stack((var, cond)), kth_distances)
            ]
            for tree, var in zip(trees_marginal_cond, data)
        ]
        count_cond = [
            tree_cond.query_ball_point(p, r=d, p=self.minkowski_p, return_length=True)
            - 1
            for p, d in zip(cond, kth_distances)
        ]

        # Compute local CMI for each data point
        local_cmi = digamma(self.k) + array(
            [
                digamma(cz + 1) - sum(digamma(c + 1) for c in counts)
                for cz, *counts in zip(count_cond, *counts_marginal_cond)
            ]
        )

        return local_cmi / log(self.base) if self.base != "e" else local_cmi



"""Functional wrappers for information estimators.

This module provides functional interfaces to calculate entropy, mutual information, and
transfer entropy.
The estimators are dynamically imported based on the estimator name provided,
saving time and memory by only importing the necessary classes.
"""

from functools import wraps

from ..utils.config import logger
from .base import EstimatorType

entropy_estimators = {
    "discrete": "infomeasure.estimators.entropy.discrete.DiscreteEntropyEstimator",
    "kernel": "infomeasure.estimators.entropy.kernel.KernelEntropyEstimator",
    "metric": "infomeasure.estimators.entropy.kozachenko_leonenko."
    "KozachenkoLeonenkoEntropyEstimator",
    "kl": "infomeasure.estimators.entropy.kozachenko_leonenko."
    "KozachenkoLeonenkoEntropyEstimator",
    "renyi": "infomeasure.estimators.entropy.renyi.RenyiEntropyEstimator",
    "ordinal": "infomeasure.estimators.entropy.ordinal.OrdinalEntropyEstimator",
    "symbolic": "infomeasure.estimators.entropy.ordinal.OrdinalEntropyEstimator",
    "permutation": "infomeasure.estimators.entropy.ordinal.OrdinalEntropyEstimator",
    "tsallis": "infomeasure.estimators.entropy.tsallis.TsallisEntropyEstimator",
}

mi_estimators = {
    "discrete": "infomeasure.estimators.mutual_information.discrete.DiscreteMIEstimator",
    "kernel": "infomeasure.estimators.mutual_information.kernel.KernelMIEstimator",
    "metric": "infomeasure.estimators.mutual_information.kraskov_stoegbauer_grassberger."
    "KSGMIEstimator",
    "ksg": "infomeasure.estimators.mutual_information.kraskov_stoegbauer_grassberger."
    "KSGMIEstimator",
    "renyi": "infomeasure.estimators.mutual_information.renyi.RenyiMIEstimator",
    "tsallis": "infomeasure.estimators.mutual_information.tsallis.TsallisMIEstimator",
    "ordinal": "infomeasure.estimators.mutual_information.ordinal.OrdinalMIEstimator",
    "symbolic": "infomeasure.estimators.mutual_information.ordinal.OrdinalMIEstimator",
    "permutation": "infomeasure.estimators.mutual_information.ordinal."
    "OrdinalMIEstimator",
}

# for M(x), use map to H(x)
mi_entropy_map = {mi_key: mi_key for mi_key in mi_estimators.keys()}
# Overwrite shorthands that differ between MI and H
mi_entropy_map["ksg"] = "kl"

cmi_estimators = {
    "discrete": "infomeasure.estimators.mutual_information.discrete.DiscreteCMIEstimator",
    "kernel": "infomeasure.estimators.mutual_information.kernel.KernelCMIEstimator",
    "metric": "infomeasure.estimators.mutual_information.kraskov_stoegbauer_grassberger."
    "KSGCMIEstimator",
    "ksg": "infomeasure.estimators.mutual_information.kraskov_stoegbauer_grassberger."
    "KSGCMIEstimator",
    "renyi": "infomeasure.estimators.mutual_information.renyi.RenyiCMIEstimator",
    "tsallis": "infomeasure.estimators.mutual_information.tsallis.TsallisCMIEstimator",
    "ordinal": "infomeasure.estimators.mutual_information.ordinal.OrdinalCMIEstimator",
    "symbolic": "infomeasure.estimators.mutual_information.ordinal.OrdinalCMIEstimator",
    "permutation": "infomeasure.estimators.mutual_information.ordinal."
    "OrdinalCMIEstimator",
}

te_estimators = {
    "discrete": "infomeasure.estimators.transfer_entropy.discrete.DiscreteTEEstimator",
    "kernel": "infomeasure.estimators.transfer_entropy.kernel.KernelTEEstimator",
    "metric": "infomeasure.estimators.transfer_entropy.kraskov_stoegbauer_grassberger."
    "KSGTEEstimator",
    "ksg": "infomeasure.estimators.transfer_entropy.kraskov_stoegbauer_grassberger."
    "KSGTEEstimator",
    "renyi": "infomeasure.estimators.transfer_entropy.renyi.RenyiTEEstimator",
    "ordinal": "infomeasure.estimators.transfer_entropy.ordinal.OrdinalTEEstimator",
    "symbolic": "infomeasure.estimators.transfer_entropy.ordinal.OrdinalTEEstimator",
    "permutation": "infomeasure.estimators.transfer_entropy.ordinal.OrdinalTEEstimator",
    "tsallis": "infomeasure.estimators.transfer_entropy.tsallis.TsallisTEEstimator",
}

cte_estimators = {
    "discrete": "infomeasure.estimators.transfer_entropy.discrete.DiscreteCTEEstimator",
    "kernel": "infomeasure.estimators.transfer_entropy.kernel.KernelCTEEstimator",
    "metric": "infomeasure.estimators.transfer_entropy.kraskov_stoegbauer_grassberger."
    "KSGCTEEstimator",
    "ksg": "infomeasure.estimators.transfer_entropy.kraskov_stoegbauer_grassberger."
    "KSGCTEEstimator",
    "renyi": "infomeasure.estimators.transfer_entropy.renyi.RenyiCTEEstimator",
    "ordinal": "infomeasure.estimators.transfer_entropy.ordinal.OrdinalCTEEstimator",
    "symbolic": "infomeasure.estimators.transfer_entropy.ordinal.OrdinalCTEEstimator",
    "permutation": "infomeasure.estimators.transfer_entropy.ordinal."
    "OrdinalCTEEstimator",
    "tsallis": "infomeasure.estimators.transfer_entropy.tsallis.TsallisCTEEstimator",
}


def _dynamic_import(class_path):
    """Dynamically import a class from a module.

    Parameters
    ----------
    class_path : str
        The path to the class to import.

    Returns
    -------
    class
        The imported class.
    """
    module_path, class_name = class_path.rsplit(".", 1)
    module = __import__(module_path, fromlist=[class_name])
    return getattr(module, class_name)


def _get_estimator(estimators, estimator_name):
    """Get the estimator class based on the estimator name.

    Parameters
    ----------
    estimators : dict
        The dictionary of available estimators.
    estimator_name : str
        The name of the estimator to get.

    Returns
    -------
    class
        The estimator class.

    Raises
    ------
    ValueError
        If the estimator is not recognized.
    """
    if (
        estimator_name is None
        or not isinstance(estimator_name, str)
        or (estimator_name.lower() not in estimators)
    ):
        available = ", ".join(estimators.keys())
        raise ValueError(
            f"Unknown estimator: {estimator_name}. Available estimators: {available}"
        )
    return _dynamic_import(estimators[estimator_name.lower()])


def get_estimator_class(measure=None, approach=None) -> EstimatorType:
    """Get estimator class based on the estimator name and approach.

    This function returns the estimator class based on the measure and approach
    provided.
    If you want an instance of an estimator, initialized with data and parameters,
    use the functional interface :func:`estimator`.

    Parameters
    ----------
    measure : str
        The measure to estimate.
        Options: ``entropy``, ``mutual_information``, ``transfer_entropy``,
        ``conditional_mutual_information``, ``conditional_transfer_entropy``.
        Aliases: ``h``, ``mi``, ``te``, ``cmi``, ``cte``.
    approach : str
        The name of the estimator to use.

    Returns
    -------
    class
        The estimator class.

    Raises
    ------
    ValueError
        If the measure is not recognized.
    ValueError
        If the approach is not recognized.
    """
    if measure is None:
        raise ValueError("The measure must be specified.")
    if measure.lower() in ["entropy", "h"]:
        return _get_estimator(entropy_estimators, approach)
    elif measure.lower() in ["mutual_information", "mi"]:
        return _get_estimator(mi_estimators, approach)
    elif measure.lower() in ["conditional_mutual_information", "cmi"]:
        return _get_estimator(cmi_estimators, approach)
    elif measure.lower() in ["transfer_entropy", "te"]:
        return _get_estimator(te_estimators, approach)
    elif measure.lower() in ["conditional_transfer_entropy", "cte"]:
        return _get_estimator(cte_estimators, approach)
    else:
        raise ValueError(
            f"Unknown measure: {measure}. Available measures: entropy, mutual_information, "
            "conditional_mutual_information, transfer_entropy, conditional_transfer_entropy."
        )


def _dynamic_estimator(measure) -> callable:
    """Decorator to dynamically inject the estimator class into the function.

    This decorator is used to inject the estimator class into the function
    based on the estimator name provided in the arguments.
    The estimator class is then used to calculate the measure.

    Parameters
    ----------
    measure : dict | [dict, dict]
        The dictionary of available estimators.
        Structure: {estimator_name: class_path}
        Or two of these, one normal and one for conditional estimators.

    Returns
    -------
    function
        The decorated function
    """

    def decorator(func):
        @wraps(func)  # This decorator updates wrapper to look like func
        def wrapper(*args, **kwargs):
            estimator_name = kwargs.get("approach")
            if estimator_name is None:
                raise ValueError("``approach`` must be provided")
            # if  `cond` is passed, it is a conditional estimator
            if isinstance(measure, str):
                # Inject EstimatorClass into kwargs
                kwargs["EstimatorClass"] = get_estimator_class(measure, estimator_name)
            elif kwargs.get("cond") is not None:
                kwargs["EstimatorClass"] = get_estimator_class(
                    measure[1], estimator_name
                )
            else:
                kwargs["EstimatorClass"] = get_estimator_class(
                    measure[0], estimator_name
                )
            return func(
                *args, **kwargs
            )  # Pass all arguments as they are, including modified kwargs

        return wrapper

    return decorator


@_dynamic_estimator("entropy")
def entropy(data, approach: str, *args, **kwargs: any):
    """Calculate the entropy using a functional interface of different estimators.

    Supports the following approaches:

    1. ``discrete``: :func:`Discrete entropy estimator. <infomeasure.estimators.entropy.discrete.DiscreteEntropyEstimator>`
    2. ``kernel``: :func:`Kernel entropy estimator. <infomeasure.estimators.entropy.kernel.KernelEntropyEstimator>`
    3. [``metric``, ``kl``]: :func:`Kozachenko-Leonenko entropy estimator. <infomeasure.estimators.entropy.kozachenko_leonenko.KozachenkoLeonenkoEntropyEstimator>`
    4. ``renyi``: :func:`Renyi entropy estimator. <infomeasure.estimators.entropy.renyi.RenyiEntropyEstimator>`
    5. [``ordinal``, ``symbolic``, ``permutation``]: :func:`Ordinal / Permutation entropy estimator. <infomeasure.estimators.entropy.ordinal.OrdinalEntropyEstimator>`
    6. ``tsallis``: :func:`Tsallis entropy estimator. <infomeasure.estimators.entropy.tsallis.TsallisEntropyEstimator>`

    Parameters
    ----------
    data : array-like
        The data used to estimate the entropy.
    approach : str
        The name of the estimator to use.
    *args: tuple
        Additional arguments to pass to the estimator.
    **kwargs: dict
        Additional keyword arguments to pass to the estimator.

    Returns
    -------
    float
        The calculated entropy.

    Raises
    ------
    ValueError
        If the estimator is not recognized.
    """
    EstimatorClass = kwargs.pop("EstimatorClass")
    return EstimatorClass(data, *args, **kwargs).result()


@_dynamic_estimator(["mi", "cmi"])
def mutual_information(
    *data,
    approach: str,
    **kwargs: any,
):
    """Calculate the mutual information using a functional interface of different
    estimators.

    Supports the following approaches:

    1. ``discrete``: :func:`Discrete mutual information estimator. <infomeasure.estimators.mutual_information.discrete.DiscreteMIEstimator>`
    2. ``kernel``: :func:`Kernel mutual information estimator. <infomeasure.estimators.mutual_information.kernel.KernelMIEstimator>`
    3. [``metric``, ``ksg``]: :func:`Kraskov-Stoegbauer-Grassberger mutual information estimator. <infomeasure.estimators.mutual_information.kraskov_stoegbauer_grassberger.KSGMIEstimator>`
    4. ``renyi``: :func:`Renyi mutual information estimator. <infomeasure.estimators.mutual_information.renyi.RenyiMIEstimator>`
    5. [``ordinal``, ``symbolic``, ``permutation``]: :func:`Ordinal mutual information estimator. <infomeasure.estimators.mutual_information.ordinal.OrdinalMIEstimator>`
    6. ``tsallis``: :func:`Tsallis mutual information estimator. <infomeasure.estimators.mutual_information.tsallis.TsallisMIEstimator>`

    Parameters
    ----------
    *data : array-like
        The data used to estimate the (conditional) mutual information.
    cond : array-like, optional
        The conditional data used to estimate the conditional mutual information.
    approach : str
        The name of the estimator to use.
    normalize : bool, optional
        If True, normalize the data before analysis. Default is False.
        Not available for the discrete estimator.
    **kwargs: dict
        Additional keyword arguments to pass to the estimator.

    Returns
    -------
    float
        The calculated mutual information.

    Raises
    ------
    ValueError
        If the estimator is not recognized.
    """
    EstimatorClass = kwargs.pop("EstimatorClass")
    if len(data) < 2:
        raise ValueError(
            "Mutual Information requires at least two variables as arguments. "
            "If needed, a conditional variable can be passed as keyword argument: "
            "`mutual_information(data1, data2, data3, ..., **kwargs)` or "
            "`mutual_information(data1, data2, ..., cond=cond_var, **kwargs)`"
        )
    return EstimatorClass(*data, **kwargs).result()


def conditional_mutual_information(*data, **kwargs: any):
    """Conditional mutual information between two variables given a third variable.

    See :func:`mutual_information <mutual_information>` for more information.
    """
    if kwargs.get("cond") is None:
        raise ValueError(
            "CMI requires a conditional variable. Pass a 'cond' keyword argument."
        )
    if len(data) < 2:
        raise ValueError(
            "CMI requires at least two variables as arguments and "
            "a conditional variable as keyword argument: "
            "`conditional_mutual_information("
            "data1, data2, ..., cond=cond_var, **kwargs)`"
        )
    return mutual_information(*data, **kwargs)


@_dynamic_estimator(["te", "cte"])
def transfer_entropy(
    *data,
    approach: str,
    **kwargs: any,
):
    """Calculate the transfer entropy using a functional interface of different estimators.

    Supports the following approaches:

    1. ``discrete``: :func:`Discrete transfer entropy estimator. <infomeasure.estimators.transfer_entropy.discrete.DiscreteTEEstimator>`
    2. ``kernel``: :func:`Kernel transfer entropy estimator. <infomeasure.estimators.transfer_entropy.kernel.KernelTEEstimator>`
    3. [``metric``, ``ksg``]: :func:`Kraskov-Stoegbauer-Grassberger transfer entropy estimator. <infomeasure.estimators.transfer_entropy.kraskov_stoegbauer_grassberger.KSGTEEstimator>`
    4. ``renyi``: :func:`Renyi transfer entropy estimator. <infomeasure.estimators.transfer_entropy.renyi.RenyiTEEstimator>`
    5. [``ordinal``, ``symbolic``, ``permutation``]: :func:`Ordinal transfer entropy estimator. <infomeasure.estimators.transfer_entropy.ordinal.OrdinalTEEstimator>`
    6. ``tsallis``: :func:`Tsallis transfer entropy estimator. <infomeasure.estimators.transfer_entropy.tsallis.TsallisTEEstimator>`

    Parameters
    ----------
    source, dest : array-like
        The source (X) and destination (Y) data used to estimate the transfer entropy.
    cond : array-like, optional
        The conditional data used to estimate the conditional transfer entropy.
    approach : str
        The name of the estimator to use.
    step_size : int
        Step size between elements for the state space reconstruction.
    src_hist_len, dest_hist_len : int
        Number of past observations to consider for the source and destination data.
    prop_time : int, optional
        Number of positions to shift the data arrays relative to each other.
        Delay/lag/shift between the variables. Default is no shift.
        Assumed time taken by info to transfer from source to destination.
        Not compatible with the ``cond`` parameter / conditional TE.
        Alternatively called ``offset``.
    *args: tuple
        Additional arguments to pass to the estimator.
    **kwargs: dict
        Additional keyword arguments to pass to the estimator.

    Returns
    -------
    float
        The calculated transfer entropy.

    Raises
    ------
    ValueError
        If the estimator is not recognized.
    """
    EstimatorClass = kwargs.pop("EstimatorClass")
    if len(data) != 2:
        raise ValueError(
            "Transfer Entropy requires two variables as arguments and if needed,"
            " the conditional data as keyword argument: "
            "`transfer_entropy(source, dest, **kwargs)` or "
            "`transfer_entropy(source, dest, cond=cond_var, **kwargs)`."
        )
    return EstimatorClass(*data, **kwargs).result()


def conditional_transfer_entropy(*data, **kwargs: any):
    """Conditional transfer entropy between two variables given a third variable.

    See :func:`transfer_entropy <transfer_entropy>` for more information.
    """
    if kwargs.get("cond") is None:
        raise ValueError(
            "CTE requires a conditional variable. Pass a 'cond' keyword argument."
        )
    if len(data) != 2:
        raise ValueError(
            "CTE requires two variables as arguments and "
            "the conditional data as keyword argument: "
            "`conditional_transfer_entropy(source, dest, cond=cond_var, **kwargs)`."
        )
    return transfer_entropy(*data, **kwargs)


def estimator(
    *data,  # *(data) for entropy, *data for mi, *(source, dest) for te
    # all arguments after this are keyword-only
    cond=None,
    measure: str = None,
    approach: str = None,
    step_size: int = 1,
    prop_time: int = 0,
    src_hist_len: int = 1,
    dest_hist_len: int = 1,
    cond_hist_len: int = 1,
    **kwargs: any,
) -> EstimatorType:
    """Get an estimator for a specific measure.

    This function provides a simple interface to get
    an :class:`Estimator <.base.Estimator>` for a specific measure.

    If you are only interested in the global result, use the functional interfaces:

    - :func:`entropy <entropy>`
    - :func:`mutual_information <mutual_information>`
    - :func:`conditional_mutual_information <conditional_mutual_information>`
    - :func:`transfer_entropy <transfer_entropy>`
    - :func:`conditional_transfer_entropy <conditional_transfer_entropy>`

    Estimators available:

    1. Entropy:
        - ``discrete``: :func:`Discrete entropy estimator. <infomeasure.estimators.entropy.discrete.DiscreteEntropyEstimator>`
        - ``kernel``: :func:`Kernel entropy estimator. <infomeasure.estimators.entropy.kernel.KernelEntropyEstimator>`
        - [``metric``, ``kl``]: :func:`Kozachenko-Leonenko entropy estimator. <infomeasure.estimators.entropy.kozachenko_leonenko.KozachenkoLeonenkoEntropyEstimator>`
        - ``renyi``: :func:`Renyi entropy estimator. <infomeasure.estimators.entropy.renyi.RenyiEntropyEstimator>`
        - [``ordinal``, ``symbolic``, ``permutation``]: :func:`Ordinal / Permutation entropy estimator. <infomeasure.estimators.entropy.ordinal.OrdinalEntropyEstimator>`
        - ``tsallis``: :func:`Tsallis entropy estimator. <infomeasure.estimators.entropy.tsallis.TsallisEntropyEstimator>`

    2. Mutual Information:
        - ``discrete``: :func:`Discrete mutual information estimator. <infomeasure.estimators.mutual_information.discrete.DiscreteMIEstimator>`
        - ``kernel``: :func:`Kernel mutual information estimator. <infomeasure.estimators.mutual_information.kernel.KernelMIEstimator>`
        - [``metric``, ``ksg``]: :func:`Kraskov-Stoegbauer-Grassberger mutual information estimator. <infomeasure.estimators.mutual_information.kraskov_stoegbauer_grassberger.KSGMIEstimator>`
        - ``renyi``: :func:`Renyi mutual information estimator. <infomeasure.estimators.mutual_information.renyi.RenyiMIEstimator>`
        - [``ordinal``, ``symbolic``, ``permutation``]: :func:`Ordinal mutual information estimator. <infomeasure.estimators.mutual_information.ordinal.OrdinalMIEstimator>`
        - ``tsallis``: :func:`Tsallis mutual information estimator. <infomeasure.estimators.mutual_information.tsallis.TsallisMIEstimator>`

    3. Transfer Entropy:
        - ``discrete``: :func:`Discrete transfer entropy estimator. <infomeasure.estimators.transfer_entropy.discrete.DiscreteTEEstimator>`
        - ``kernel``: :func:`Kernel transfer entropy estimator. <infomeasure.estimators.transfer_entropy.kernel.KernelTEEstimator>`
        - [``metric``, ``ksg``]: :func:`Kraskov-Stoegbauer-Grassberger transfer entropy estimator. <infomeasure.estimators.transfer_entropy.kraskov_stoegbauer_grassberger.KSGTEEstimator>`
        - ``renyi``: :func:`Renyi transfer entropy estimator. <infomeasure.estimators.transfer_entropy.renyi.RenyiTEEstimator>`
        - [``ordinal``, ``symbolic``, ``permutation``]: :func:`Ordinal transfer entropy estimator. <infomeasure.estimators.transfer_entropy.ordinal.OrdinalTEEstimator>`
        - ``tsallis``: :func:`Tsallis transfer entropy estimator. <infomeasure.estimators.transfer_entropy.tsallis.TsallisTEEstimator>`

    Parameters
    ----------
    *data :
        The data used to estimate the measure.
        For entropy: a single array-like data. A tuple of data for joint entropy.
        For mutual information: arbitrary number of array-like data.
        For transfer entropy: two array-like data. Source and destination.
    cond : array-like, optional
        Only if the measure is conditional transfer entropy.
    measure : str
        The measure to estimate.
        Options: ``entropy``, ``mutual_information``, ``transfer_entropy``,
        ``conditional_mutual_information``
        ``conditional_transfer_entropy``;
        aliases: ``h``, ``mi``, ``te``, ``cmi``, ``cte``.
    approach : str
        The name of the estimator to use.
        Find the available estimators in the docstring of this function.
    *args: tuple
        Additional arguments to pass to the estimator.
    **kwargs: dict
        Additional keyword arguments to pass to the estimator.

    Returns
    -------
    Estimator
        The estimator instance.

    Raises
    ------
    ValueError
        If the measure is not recognized.
    """
    if measure is None:
        raise ValueError("``measure`` is required.")
    elif measure.lower() in ["entropy", "h"]:
        if len(data) == 0:
            raise ValueError("``data`` is required for entropy estimation.")
        if cond is not None:
            raise ValueError("``cond`` is not required for entropy estimation.")
        if len(data) != 1:
            raise ValueError(
                "Exactly one data array is required for entropy estimation. "
                f"Got {len(data)}. "
                "To signal that you want to compute joint entropy, "
                "pass your data in a tuple, e.g., (data1, data2)."
            )
        EstimatorClass = _get_estimator(entropy_estimators, approach)
        return EstimatorClass(data[0], **kwargs)
    elif measure.lower() in [
        "mutual_information",
        "mi",
        "conditional_mutual_information",
        "cmi",
    ]:
        if (
            measure.lower() in ["cmi", "conditional_mutual_information"]
            and cond is None
        ):
            raise ValueError(
                "No conditional data was provided for conditional mutual information"
                "estimation. Pass ``cond`` to specify the conditional data."
            )
        if len(data) == 0:
            raise ValueError("No data was provided for mutual information estimation.")
        if len(data) == 1:
            logger.warning(
                "Only one data array provided for mutual information estimation. "
                "Using normal entropy estimator."
            )
            EstimatorClass = _get_estimator(
                entropy_estimators, mi_entropy_map[approach]
            )
            return EstimatorClass(data[0], **kwargs)
        if cond is not None:
            EstimatorClass = _get_estimator(cmi_estimators, approach)
            return EstimatorClass(*data, cond=cond, **kwargs)
        else:
            EstimatorClass = _get_estimator(mi_estimators, approach)
            return EstimatorClass(*data, **kwargs)
    elif measure.lower() in [
        "transfer_entropy",
        "te",
        "conditional_transfer_entropy",
        "cte",
    ]:
        if measure.lower() in ["cte", "conditional_transfer_entropy"] and cond is None:
            raise ValueError(
                "No conditional data was provided for conditional transfer entropy "
                "estimation. Pass ``cond`` to specify the conditional data."
            )
        if len(data) != 2:
            raise ValueError(
                "Exactly two data arrays are required for transfer entropy estimation."
            )
        if cond is not None:
            EstimatorClass = _get_estimator(cte_estimators, approach)
            return EstimatorClass(
                *data,
                cond=cond,
                prop_time=prop_time,
                src_hist_len=src_hist_len,
                dest_hist_len=dest_hist_len,
                cond_hist_len=cond_hist_len,
                step_size=step_size,
                **kwargs,
            )
        else:
            EstimatorClass = _get_estimator(te_estimators, approach)
            return EstimatorClass(
                *data,
                prop_time=prop_time,
                src_hist_len=src_hist_len,
                dest_hist_len=dest_hist_len,
                step_size=step_size,
                **kwargs,
            )
    else:
        raise ValueError(f"Unknown measure: {measure}")



"""Module for the Tsallis mutual information estimator."""

from abc import ABC

from numpy import issubdtype, integer

from ..base import (
    PValueMixin,
    MutualInformationEstimator,
    ConditionalMutualInformationEstimator,
)
from ..entropy.tsallis import TsallisEntropyEstimator
from ... import Config
from ...utils.types import LogBaseType


class BaseTsallisMIEstimator(ABC):
    r"""Base class for Tsallis mutual information estimators.

    Attributes
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the (conditional) mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
        For conditional mutual information, only two data arrays are allowed.
    cond : array-like, optional
        The conditional data used to estimate the conditional mutual information.
    k : int
        The number of nearest neighbors used in the estimation.
    q : float
        The Tsallis parameter, order or exponent.
        Sometimes denoted as :math:`q`, analogous to the Rényi parameter :math:`\alpha`.
    noise_level : float
        The standard deviation of the Gaussian noise to add to the data to avoid
        issues with zero distances.
    offset : int, optional
        Number of positions to shift the data arrays relative to each other.
        Delay/lag/shift between the variables. Default is no shift.
        Not compatible with the ``cond`` parameter / conditional MI.
    normalize : bool, optional
        If True, normalize the data before analysis.
    """

    def __init__(
        self,
        *data,
        cond=None,
        k: int = 4,
        q: float | int = None,
        noise_level: float = 1e-8,
        offset: int = 0,
        normalize: bool = False,
        base: LogBaseType = Config.get("base"),
    ):
        r"""Initialize the BaseTsallisMIEstimator.

        Parameters
        ----------
        *data : array-like, shape (n_samples,)
            The data used to estimate the (conditional) mutual information.
            You can pass an arbitrary number of data arrays as positional arguments.
            For conditional mutual information, only two data arrays are allowed.
        cond : array-like, optional
            The conditional data used to estimate the conditional mutual information.
        k : int
            The number of nearest neighbors to consider.
        q : float | int
            The Tsallis parameter, order or exponent.
            Sometimes denoted as :math:`q`,
            analogous to the Rényi parameter :math:`\alpha`.
        noise_level : float
            The standard deviation of the Gaussian noise to add to the data to avoid
            issues with zero distances.
        normalize
            If True, normalize the data before analysis.
        offset : int, optional
            Number of positions to shift the X and Y data arrays relative to each other.
            Delay/lag/shift between the variables. Default is no shift.
            Not compatible with the ``cond`` parameter / conditional MI.

        Raises
        ------
        ValueError
            If the Tsallis parameter is not a positive number.
        ValueError
            If the number of nearest neighbors is not a positive integer.
        """
        if cond is None:
            super().__init__(*data, offset=offset, normalize=normalize, base=base)
        else:
            super().__init__(
                *data, cond=cond, offset=offset, normalize=normalize, base=base
            )

        if not isinstance(q, (int, float)) or q <= 0:
            raise ValueError("The Tsallis parameter ``q`` must be a positive number.")
        if not issubdtype(type(k), integer) or k <= 0:
            raise ValueError(
                "The number of nearest neighbors must be a positive integer."
            )

        self.k = k
        self.q = q
        self.noise_level = noise_level


class TsallisMIEstimator(
    BaseTsallisMIEstimator, PValueMixin, MutualInformationEstimator
):
    r"""Estimator for the Tsallis mutual information.

    Attributes
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
    k : int
        The number of nearest neighbors used in the estimation.
    q : float
        The Tsallis parameter, order or exponent.
        Sometimes denoted as :math:`q`, analogous to the Rényi parameter :math:`\alpha`.
    noise_level : float
        The standard deviation of the Gaussian noise to add to the data to avoid
        issues with zero distances.
    offset : int, optional
        Number of positions to shift the data arrays relative to each other.
        Delay/lag/shift between the variables. Default is no shift.
    normalize : bool, optional
        If True, normalize the data before analysis.

    Notes
    -----
    In the :math:`q \to 1` limit, the Jackson sum (q-additivity) reduces to
    ordinary summation,
    and the Tallis entropy reduces to Shannon Entropy.
    This class of entropy measure is in particularly useful in the study in connection
    with long-range correlated systems and with non-equilibrium phenomena.
    """

    def _calculate(self):
        """Calculate the mutual information of the data.

        Returns
        -------
        float
            Tsallis mutual information of the data.
        """

        return self._generic_mi_from_entropy(
            estimator=TsallisEntropyEstimator,
            noise_level=self.noise_level,
            kwargs={"q": self.q, "k": self.k, "base": self.base},
        )


class TsallisCMIEstimator(
    BaseTsallisMIEstimator, ConditionalMutualInformationEstimator
):
    r"""Estimator for the conditional Tsallis mutual information.

    Parameters
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the conditional mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
    cond : array-like
        The conditional data used to estimate the conditional mutual information.
    k : int
        The number of nearest neighbors to consider.
    q : float | int
        The Tsallis parameter, order or exponent.
        Sometimes denoted as :math:`q`,
        analogous to the Rényi parameter :math:`\alpha`.
    noise_level : float
        The standard deviation of the Gaussian noise to add to the data to avoid
        issues with zero distances.
    normalize
        If True, normalize the data before analysis.

    Notes
    -----
    In the :math:`q \to 1` limit, the Jackson sum (q-additivity) reduces to
    ordinary summation,
    and the Tallis entropy reduces to Shannon Entropy.
    This class of entropy measure is in particularly useful in the study in connection
    with long-range correlated systems and with non-equilibrium phenomena.
    """

    def _calculate(self):
        """Calculate the conditional mutual information of the data.

        Returns
        -------
        float
            Tsallis conditional mutual information of the data.
        """
        return self._generic_cmi_from_entropy(
            estimator=TsallisEntropyEstimator,
            noise_level=self.noise_level,
            kwargs=dict(q=self.q, k=self.k, base=self.base),
        )



"""Module for the discrete mutual information estimator."""

from abc import ABC

from numpy import ndarray

from ..base import (
    PValueMixin,
    MutualInformationEstimator,
    ConditionalMutualInformationEstimator,
)
from ..utils.discrete_interaction_information import (
    mutual_information_global,
    mutual_information_local,
    conditional_mutual_information_global,
    conditional_mutual_information_local,
)
from ... import Config
from ...utils.config import logger
from ...utils.types import LogBaseType


class BaseDiscreteMIEstimator(ABC):
    """Base class for discrete mutual information estimators.

    Attributes
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the (conditional) mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
        For conditional mutual information, only two data arrays are allowed.
    cond : array-like, optional
        The conditional data used to estimate the conditional mutual information.
    offset : int, optional
        Number of positions to shift the data arrays relative to each other.
        Delay/lag/shift between the variables. Default is no shift.
        Not compatible with the ``cond`` parameter / conditional MI.
    """

    def __init__(
        self,
        *data,
        cond=None,
        offset: int = 0,
        base: LogBaseType = Config.get("base"),
    ):
        """Initialize the BaseDiscreteMIEstimator.

        Parameters
        ----------
        *data : array-like, shape (n_samples,)
            The data used to estimate the (conditional) mutual information.
            You can pass an arbitrary number of data arrays as positional arguments.
            For conditional mutual information, only two data arrays are allowed.
        cond : array-like, optional
            The conditional data used to estimate the conditional mutual information.
        offset : int, optional
            Number of positions to shift the X and Y data arrays relative to each other.
            Delay/lag/shift between the variables. Default is no shift.
            Not compatible with the ``cond`` parameter / conditional MI.
        """
        self.data: tuple[ndarray] = None
        if cond is None:
            super().__init__(*data, offset=offset, normalize=False, base=base)
        else:
            super().__init__(
                *data, cond=cond, offset=offset, normalize=False, base=base
            )
        if any(var.dtype.kind == "f" for var in self.data):
            logger.warning(
                "The data looks like a float array ("
                f"{[var.dtype for var in self.data]}). "
                "Make sure it is properly symbolized or discretized "
                "for the mutual information estimation."
            )
        if hasattr(self, "cond") and self.cond.dtype.kind == "f":
            logger.warning(
                "The conditional data looks like a float array ("
                f"{self.cond.dtype}). "
                "Make sure it is properly symbolized or discretized "
                "for the conditional mutual information estimation."
            )


class DiscreteMIEstimator(
    BaseDiscreteMIEstimator, PValueMixin, MutualInformationEstimator
):
    """Estimator for the discrete mutual information.

    Attributes
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
    offset : int, optional
        Number of positions to shift the data arrays relative to each other.
        Delay/lag/shift between the variables. Default is no shift.
    """

    def _calculate(self):
        """Calculate the mutual information of the data.

        Returns
        -------
        float
            The calculated mutual information.
        """
        return mutual_information_global(*self.data, log_func=self._log_base)

    def _extract_local_values(self) -> ndarray[float]:
        """Separately calculate the local values.

        Returns
        -------
        ndarray[float]
            The calculated local values of mi.
        """
        return mutual_information_local(*self.data, log_func=self._log_base)


class DiscreteCMIEstimator(
    BaseDiscreteMIEstimator, ConditionalMutualInformationEstimator
):
    """Estimator for the discrete conditional mutual information.

    Attributes
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the conditional mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
    cond : array-like
        The conditional data used to estimate the conditional mutual information.
    """

    def _calculate(self):
        """Calculate the conditional mutual information of the data.

        Returns
        -------
        float
            The calculated conditional mutual information.
        """
        return conditional_mutual_information_global(
            *self.data, cond=self.cond, log_func=self._log_base
        )

    def _extract_local_values(self) -> ndarray:
        """Separately calculate the local values.

        Returns
        -------
        ndarray[float]
            The calculated local values of cmi.
        """
        return conditional_mutual_information_local(
            *self.data, cond=self.cond, log_func=self._log_base
        )



"""Helper functions for exponential family distributions.

Rényi entropy and Tsallis entropy are special cases of the more general
family of exponential family distributions. This module provides helper
functions for these distributions.
"""

from numpy import pi, mean as np_mean, exp as np_exp
from scipy.spatial import KDTree
from scipy.special import gamma, digamma


def calculate_common_entropy_components(data, k):
    """Calculate common components for entropy estimators.

    Parameters
    ----------
    data : array-like
        The data used to estimate the entropy.
    k : int
        The number of nearest neighbors used in the estimation.

    Returns
    -------
    tuple
        Volume of the unit ball, k-th nearest neighbor distances,
        number of data points, and dimensionality of the data.

    Raises
    ------
    ValueError
        If the parameter ``k`` is selected too large.
    """
    N, m = data.shape

    if k >= N:
        raise ValueError(
            "The number of nearest neighbors must be smaller "
            "than the number of data points."
        )

    # Volume of the unit ball in m-dimensional space
    V_m = pi ** (m / 2) / gamma(m / 2 + 1)

    # Build k-d tree for nearest neighbor search
    tree = KDTree(data)

    # Get the k-th nearest neighbor distances
    rho_k = tree.query(data, k=k + 1)[0][
        :, k
    ]  # k+1 because the point itself is included

    return V_m, rho_k, N, m


def exponential_family_iq(k, q, V_m, rho_k, N, m):
    r"""Calculate the :math:`I_q` of the exponential family distribution.

    Parameters
    ----------
    k : int
        The number of nearest neighbors used in the estimation.
    q : float | int
        The Rényi or Tsallis parameter, order or exponent.
        Sometimes denoted as :math:`\alpha` or :math:`q`.
        Should not be 1.
    V_m : float
        Volume of the unit ball in m-dimensional space.
    rho_k : array-like
        The k-th nearest neighbor distances.
    N : int
        Number of data points.
    m : int
        Dimensionality of the data.

    Returns
    -------
    float
        The :math:`I_q` of the exponential family distribution
    """
    C_k = (gamma(k) / gamma(k + 1 - q)) ** (1 / (1 - q))
    zeta_N_i_k = (N - 1) * C_k * V_m * rho_k**m
    return np_mean(zeta_N_i_k ** (1 - q))


def exponential_family_i1(k, V_m, rho_k, N, m, log_base_func):
    r"""Calculate the :math:`I_1` of the exponential family distribution.

    When :math:`q = 1`, the exponential family distribution reduces to the
    Shannon entropy.

    Parameters
    ----------
    k : int
        The number of nearest neighbors used in the estimation.
    V_m : float
        Volume of the unit ball in m-dimensional space.
    rho_k : array-like
        The k-th nearest neighbor distances.
    N : int
        Number of data points.
    m : int
        Dimensionality of the data.
    log_base_func : callable
        The logarithm function to use for the calculation with the chosen base.

    Returns
    -------
    float
        The :math:`I_1` of the exponential family distribution
    """
    zeta_N_i_k = (N - 1) * np_exp(-digamma(k)) * V_m * rho_k**m
    return np_mean(log_base_func(zeta_N_i_k))
